/*
* Copyright (c) 2016, Intel Corporation
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* - Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
* - Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
* - Neither the name of Intel Corporation nor the names of its contributors
*   may be used to endorse or promote products derived from this software
*   without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/

#pragma once

#include <vector>
#include <algorithm>
#include "../common/math/vec2.h"
#include "../common/math/vec3.h"
#include "../common/math/vec3fa.h"

#include "defines.h"

#ifdef ENABLE_RANGE_CHECK
#define RANGE_CHECK(X, A, B) { if (X < A || X > B) printf("[%s, L:%04d], Range check failed, %d != [%d, %d]\n", __FILE__, __LINE__, X, A, B); }
#else
#define RANGE_CHECK(X, A, B)
#endif

namespace embree
{	
	/* A simple mipmapped cache implementation */
	template<class T> class MipCache
	{
	public:
		struct Element
		{
			T		m_data;		// Templated datatype
			LONG	m_shade;	// Flag to indicate if element has been updated
			LONG	m_initialized;
		};

		enum MutexStatus
		{
			ELEMENT_FREE = 0,
			ELEMENT_SHADED = 1,
			ELEMENT_BUSY = 2
		};

		MipCache(int baseResolution, int mipLevels, int numFaces)
			: m_resolution(baseResolution)
			, m_numMipLevels(mipLevels)
		{
			m_shade.resize(numFaces);
			for (int i = 0; i < numFaces; ++i)
			{
				MipLevel *mipData = new MipLevel[mipLevels];
				for (int mip = 0; mip < mipLevels; ++mip)
				{
					int mipRes = m_resolution >> mip;
					assert(mipRes > 0);
					mipData[mip].m_data = new Element[mipRes*mipRes];
				}
				m_mipData.push_back(mipData);
				m_shade[i] = false;;
			}
			ClearFull();
		}

		~MipCache()
		{
			for (size_t i = 0; i < (int)m_mipData.size(); ++i)
			{
				for (int mip = 0; mip < m_numMipLevels; ++mip)
				{
					if (m_mipData[i][mip].m_data != nullptr)
						delete[] m_mipData[i][mip].m_data;
					m_mipData[i][mip].m_data = nullptr;
				}
				delete[] m_mipData[i];
			}
			m_mipData.clear();
		}

		void ClearFull() {

#pragma omp parallel for
			for (int i = 0; i < (int)m_mipData.size(); ++i)
			{
				for (int mip = 0; mip < m_numMipLevels; ++mip)
				{
					m_mipData[i][mip].m_shade = 0;
					int mipRes = m_resolution >> mip;
					for (int idx = 0; idx < mipRes*mipRes; ++idx)
					{
						m_mipData[i][mip].m_data[idx].m_shade = 0;
						m_mipData[i][mip].m_data[idx].m_initialized = 0;
					}
				}
				m_shade[i] = 0;
			}
		}

		void ClearFast() {
			
#pragma omp parallel for
			for (int i = 0; i < (int)m_mipData.size(); ++i)
			{
				if (m_shade[i])
				{
					for (int mip = 0; mip < m_numMipLevels; ++mip)
					{
						if (!m_mipData[i][mip].m_shade)
							continue;
						m_mipData[i][mip].m_shade = 0;

						int mipRes = m_resolution >> mip;
						for (int idx = 0; idx < mipRes*mipRes; ++idx)
							m_mipData[i][mip].m_data[idx].m_shade = 0;

					}
					m_shade[i] = 0;
				}
			}
		}

		bool GetShade(int faceID) const
		{ 
			RANGE_CHECK(faceID, 0, (int)m_mipData.size());
			return m_shade[faceID];
		}
		void SetShade(int faceID, bool val)
		{
			RANGE_CHECK(faceID, 0, (int)m_mipData.size());
			m_shade[faceID] = val;
		}
		
		bool GetMipShade(int faceID, int mipLevel) const
		{
			RANGE_CHECK(faceID, 0, (int)m_mipData.size());
			RANGE_CHECK(mipLevel, 0, m_numMipLevels - 1);
			return m_mipData[faceID][mipLevel].m_shade;
		}
		void SetMipShade(int faceID, int mipLevel, bool val)
		{
			RANGE_CHECK(faceID, 0, (int)m_mipData.size());
			RANGE_CHECK(mipLevel, 0, m_numMipLevels - 1);
			if (val)
				m_shade[faceID] = true;
			m_mipData[faceID][mipLevel].m_shade = val ? 1 : 0;
		}

		bool GetTexelShade(int faceID, int x, int y, int mipLevel) const
		{
			RANGE_CHECK(faceID, 0, (int)m_mipData.size());
			RANGE_CHECK(x, 0, m_resolution - 1);
			RANGE_CHECK(y, 0, m_resolution - 1);
			RANGE_CHECK(mipLevel, 0, m_numMipLevels - 1);
			int mipIdx = (x >> mipLevel) + (y >> mipLevel)*(m_resolution >> mipLevel);
			return m_mipData[faceID][mipLevel].m_data[mipIdx].m_shade;
		}

		void SetTexelShade(int faceID, int x, int y, int mipLevel, bool val)
		{
			RANGE_CHECK(faceID, 0, (int)m_mipData.size());
			RANGE_CHECK(x, 0, m_resolution - 1);
			RANGE_CHECK(y, 0, m_resolution - 1);
			RANGE_CHECK(mipLevel, 0, m_numMipLevels - 1);
			int mipIdx = (x >> mipLevel) + (y >> mipLevel)*(m_resolution >> mipLevel);
			if (val) 
			{
				m_mipData[faceID][mipLevel].m_shade = true;
				m_shade[faceID] = true;
			}
			m_mipData[faceID][mipLevel].m_data[mipIdx].m_shade = val ? 1 : 0;
		}

		Element &GetTexelElement(int faceID, int x, int y, int mipLevel)
		{
			RANGE_CHECK(faceID, 0, (int)m_mipData.size());
			RANGE_CHECK(x, 0, m_resolution - 1);
			RANGE_CHECK(y, 0, m_resolution - 1);
			RANGE_CHECK(mipLevel, 0, m_numMipLevels - 1);
			int mipIdx = (x >> mipLevel) + (y >> mipLevel)*(m_resolution >> mipLevel);
			return m_mipData[faceID][mipLevel].m_data[mipIdx];
		}

		inline bool RequiresProcessingSync(int faceID, int x, int y, int mipLevel)
		{
			RANGE_CHECK(faceID, 0, (int)m_mipData.size());
			RANGE_CHECK(x, 0, m_resolution - 1);
			RANGE_CHECK(y, 0, m_resolution - 1);
			RANGE_CHECK(mipLevel, 0, m_numMipLevels - 1);
			int mipIdx = (x >> mipLevel) + (y >> mipLevel)*(m_resolution >> mipLevel);
					
			// Early bail if shaded
			if (m_mipData[faceID][mipLevel].m_data[mipIdx].m_shade == ELEMENT_SHADED)
				return false;
			
			MutexStatus ms = (MutexStatus)InterlockedCompareExchange(&m_mipData[faceID][mipLevel].m_data[mipIdx].m_shade, ELEMENT_BUSY, ELEMENT_FREE);
			while (ms == ELEMENT_BUSY)
			{
				ms = (MutexStatus)InterlockedCompareExchange(&m_mipData[faceID][mipLevel].m_data[mipIdx].m_shade, ELEMENT_BUSY, ELEMENT_FREE);
			}
			return ms == ELEMENT_FREE;
		}

		void SignalProcessed(int faceID, int x, int y, int mipLevel)
		{
			RANGE_CHECK(faceID, 0, (int)m_mipData.size());
			RANGE_CHECK(x, 0, m_resolution - 1);
			RANGE_CHECK(y, 0, m_resolution - 1);
			RANGE_CHECK(mipLevel, 0, m_numMipLevels - 1);
			int mipIdx = (x >> mipLevel) + (y >> mipLevel)*(m_resolution >> mipLevel);
			InterlockedCompareExchange(&m_mipData[faceID][mipLevel].m_data[mipIdx].m_shade, ELEMENT_SHADED, ELEMENT_BUSY);
		}

	protected:

		struct MipLevel
		{
			Element *m_data;
			LONG	m_shade;
		};

		int						m_resolution;
		int						m_numMipLevels;
		std::vector<LONG>		m_shade;
		std::vector<MipLevel*>	m_mipData;
	};
}
