/*
* Copyright (c) 2016, Intel Corporation
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* - Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
* - Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
* - Neither the name of Intel Corporation nor the names of its contributors
*   may be used to endorse or promote products derived from this software
*   without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/
#include <mutex>
#include <unordered_map>

#include "patchtexture.h"
//#include "../common/sys/alloc.h"

#include "renderer.h"

namespace embree
{
	static const int FILTER_MAX_SIZE = 32;

	PatchTexture::PatchTexture(const int geomID, const int numFaces, const int shadingRes,
							   const std::vector<int>& face_neighbors, const std::vector<int>& edge_neighbors,
							   const std::vector<int>& edges_per_face, const Renderer* pRenderer)
		: m_numFaces(numFaces)
		, m_shadingRes(shadingRes)
		, m_geomID(geomID)
		, m_pRenderer(pRenderer)
		, m_accumulate(false)
	{

		// Create mipmap pyramid
		m_numMipLevels = 0;
		for (int mipRes = m_shadingRes; mipRes > 0; mipRes /= 2)
		{
			m_numMipLevels++;
		}

		m_shadingCache = new ShadingCache(m_shadingRes, m_numMipLevels, m_numFaces);
		m_filterCache = new FilteringCache(m_shadingRes, m_numMipLevels, m_numFaces);

		// Allocate local shading buffer (one set of buffers are reused for all faces)
		const int maxFilterWidth = m_shadingRes + 2 * FILTER_MAX_SIZE;

		// Setup neighbor information
		const int N = numFaces * 4;
		m_face_neighbors = new int[N];
		m_edge_neighbors = new int[N];
		m_edges_per_face = new int[numFaces];

		for (int i = 0; i < N; ++i)
		{
			m_face_neighbors[i] = face_neighbors[i];
			m_edge_neighbors[i] = edge_neighbors[i];
		}

		for (int i = 0; i < numFaces; ++i)
		{
			m_edges_per_face[i] = edges_per_face[i];
		}

		ClearFull();
	}

	PatchTexture::~PatchTexture()
	{
		SAFE_DELETE(m_shadingCache);
		SAFE_DELETE(m_filterCache);
		SAFE_DELETE_ARRAY(m_face_neighbors);
		SAFE_DELETE_ARRAY(m_edge_neighbors);
		SAFE_DELETE_ARRAY(m_edges_per_face);
	}

	void PatchTexture::FindSample(unsigned int faceID, int u, int v, unsigned int &outFaceID, int &outU, int &outV, unsigned int depth) const
	{
		RANGE_CHECK(faceID, 0, m_numFaces - 1);
		if (depth > 1)
		{
			outU = max(0, min(m_shadingRes - 1, u));
			outV = max(0, min(m_shadingRes - 1, v));
			outFaceID = faceID;
			return;
		}
		
		int idxN = faceID;
		int r = 0;
		int remap_u, remap_v;

		if (u < 0)
		{
			idxN = GetFaceNeighbor(faceID, 3);
			r = GetRotation(faceID, 3);
			remap_u = m_shadingRes + u;
			remap_v = v;
		}
		else if (u > m_shadingRes - 1)
		{
			idxN = GetFaceNeighbor(faceID, 1);
			r = GetRotation(faceID, 1);
			remap_u = u - m_shadingRes;
			remap_v = v;
		}
		else if (v < 0)
		{
			idxN = GetFaceNeighbor(faceID, 0);
			r = GetRotation(faceID, 0);
			remap_u = u;
			remap_v = m_shadingRes + v;
		}
		else if (v > m_shadingRes - 1)
		{
			idxN = GetFaceNeighbor(faceID, 2);
			r = GetRotation(faceID, 2);
			remap_u = u;
			remap_v = v - m_shadingRes;
		}
		else
		{
			outU = u;
			outV = v;
			outFaceID = faceID;
			return;
		}

		if (idxN != -1)
		{
			Rotate(r, remap_u, remap_v);
			FindSample(idxN, remap_u, remap_v, outFaceID, outU, outV, depth + 1);
			return;
		}

		outU = max(0, min(m_shadingRes - 1, u));
		outV = max(0, min(m_shadingRes - 1, v));
		outFaceID = faceID;
	}

	Vec3fa PatchTexture::GetFilteredValue(const int filterRadius, const float mipBias, int primID, float u, float v, const float time, const Vec2f &dUVdX, const Vec2f &dUVdY)
	{
		Vec2f ddx = dUVdX * (float)m_shadingRes;
		Vec2f ddy = dUVdY * (float)m_shadingRes;

#ifdef CLAMP_MIP_TO_FILTER
		// Clamp the miplevel to ensure that we have at least as many texels available as specified by the filter footprint. 
		// We only filter using the direct 1-ring neighbours of a patch. For example, if our filter footprint is 9x9 pixels 
		// (radius = 4) we clamp to the 4x4 texel mip-level to ensure that the filter never extends beyond the 1-ring.
		int mipMax = (int)ceil(log2f(m_shadingRes)) - max(0, (int)ceil(log2f(filterRadius)));
		float mip = min(float(mipMax), ComputeMIPf(ddx, ddy, mipBias));
#else
		int mip = ComputeMIP(ddx, ddy, mipBias);
#endif
		return FilterTexel(primID, u, v, mip, time, filterRadius);
	}

	bool PatchTexture::GetShadedSampleAsync(int faceID, int x, int y, int mipLevel, float time, FatSample &s)
	{
		RANGE_CHECK(faceID, 0, m_numFaces - 1);

		// Check if the texel resides in the cache
		ShadingCache::Element &e = m_shadingCache->GetTexelElement(faceID, x, y, mipLevel);
		if (e.m_shade != ShadingCache::ELEMENT_SHADED)
		{
			// Try to get exclusive access to shade the texel. If the lock fails, return and let the callee handle the conflict
			if (InterlockedCompareExchange(&e.m_shade, ShadingCache::ELEMENT_BUSY, ShadingCache::ELEMENT_FREE) == ShadingCache::ELEMENT_BUSY)
				return false;

			// Texel was successfully locked: execute shade callback
			FatSample shaded;
			m_pRenderer->texelCallback(m_geomID, faceID, x >> mipLevel, y >> mipLevel, m_shadingRes >> mipLevel, time, shaded);
			if (m_accumulate && e.m_initialized != 0)
				e.m_data.accumulate(shaded);	// Accumulate data into the cache (e.g. temporal re-use)
			else
			{
				e.m_data = shaded; // Overwrite the texel with the shaded value
				e.m_initialized = 1;
			}

			// Flag texel as shaded
			m_shadingCache->SetMipShade(faceID, mipLevel, true);
			m_shadingCache->SetShade(faceID, true);
			e.m_shade = ShadingCache::ELEMENT_SHADED;
		}

		s = e.m_data;
		return true;
	}

	Vec3fa PatchTexture::BoxFilterWithNormalThreshold(FilterWindow &window, const FatSample &centerTexel)
	{
		Vec3fa res(0);
		for (int lv = window.vs; lv <= window.ve; ++lv)
		{
			for (int lu = window.us; lu <= window.ue; ++lu)
			{
				int tapIdx = (lu - window.us) + (lv - window.vs)*(window.ue - window.us + 1);
				FatSample &currentTexel = window.filterTexels[tapIdx];

				float v = dot(centerTexel.normal, currentTexel.normal);
				if (v < NORMAL_ANGULAR_THRESHOLD) 
					continue;

				res += Vec3fa(currentTexel.color, 1.0f);
			}
		}

		return fabs(res.w) > 0.00001 ? res / res.w : res;
	}

	Vec3fa PatchTexture::FilterTexel(int faceID, const float u, const float v, float mipLevelf, float time, float filterRadius)
	{
		int mipLevel = int(mipLevelf);

		float fu = (u*m_shadingRes);
		float fv = (v*m_shadingRes);
		float mipScale = 1.0f / (float)(1 << mipLevel);

		int iu = min((int)fu, m_shadingRes - 1);
		int iv = min((int)fv, m_shadingRes - 1);

		// Find cached filtered value, if any
		Vec3f filtered;
		bool filter = m_filterCache->RequiresProcessingSync(faceID, iu, iv, mipLevel);
		FilteringCache::Element &e = m_filterCache->GetTexelElement(faceID, iu, iv, mipLevel);
		if (filter)
		{
			int us = (int)(fu*mipScale - filterRadius);
			int vs = (int)(fv*mipScale - filterRadius);
			int ue = (int)(fu*mipScale + filterRadius);
			int ve = (int)(fv*mipScale + filterRadius);

			FatSample filterTexels[FILTER_MAX_SIZE * FILTER_MAX_SIZE];
			int numTexels = max(0, (ue - us + 1)*(ve - vs + 1)), processedTexels = 0;

			// Spin lock until all texels in the filter footprint have been shaded. 
			while (processedTexels < numTexels) 
			{
				processedTexels = 0;
				for (int lv = vs; lv <= ve; ++lv)
				{
					for (int lu = us; lu <= ue; ++lu)
					{
						int outU, outV;
						unsigned int outFaceID;
						FindSample(faceID, lu << mipLevel, lv << mipLevel, outFaceID, outU, outV);

						int tapIdx = (lu - us) + (lv - vs)*(ue - us + 1);
						
						// Query the cache for a shaded texel:
						//     1. If the cached texel contains the shaded result it's returned immediately
						//     2. If the cached texel is empty, this thread will run a callback function shading the texel
						//     3. If the cached texel is shaded by another thread, the function returns false
						bool isShaded = GetShadedSampleAsync(outFaceID, outU, outV, mipLevel, time, filterTexels[tapIdx]);
						
						// Count the number of texels successfully shaded (not currently processed by another thread)
						if (isShaded)
							processedTexels++;
					}
				}
			}

			// Execute filter
			FilterWindow window = {us, vs, ue, ve, iu, iv, filterTexels};

			int ciu = iu >> mipLevel;
			int civ = iv >> mipLevel;

			int tapIdx = (ciu - us) + (civ - vs)*(ue - us + 1);
			FatSample &centerTexel = filterTexels[tapIdx];

			filtered = BoxFilterWithNormalThreshold(window, centerTexel);

			e.m_data = filtered;
			m_filterCache->SetMipShade(faceID, mipLevel, true);
			m_filterCache->SetShade(faceID, true);
		}
		else
			filtered = e.m_data;
		m_filterCache->SignalProcessed(faceID, iu, iv, mipLevel);

		return filtered;
	}

	/* Compute valence for a face vertex*/
	int PatchTexture::ComputeValence(int faceID, int vertex) const
	{
		// Walk clockwise along faces until we return to the same face or hit an edge

		int numEdges = m_edges_per_face[faceID];

		int valence = 0;
		int side = (vertex + numEdges-1) % numEdges;
		int newFace = faceID;

		do {
			int edge = GetEdgeNeighbor(newFace, side);
			newFace = GetFaceNeighbor(newFace, side);

			int ne = newFace != -1 ? m_edges_per_face[newFace] : 4;
			side = (edge + ne-1) % ne; // previous edge
			valence++;
		} while (newFace != -1 && newFace != faceID);

		if (newFace == -1)
		{
			// Walk ccw along faces until we hit an edge
			side = vertex % numEdges;
			newFace = GetFaceNeighbor(faceID, side);
			side = GetEdgeNeighbor(faceID, side);
			while (newFace != -1 && newFace != faceID)
			{
				int ne = m_edges_per_face[newFace];
				side = (side +1) % ne;
				newFace = GetFaceNeighbor(newFace, side);
				side = GetEdgeNeighbor(newFace, side);
				valence++;
			}
		}
		return valence;
	}

	// Rotate u,v parameterization, when moving to an adjacent (p-tex) patch
	void PatchTexture::Rotate(const int r, int &x, int &y) const
	{
		int xr = x;
		int yr = y;

		if (r == -1 || r == 3)
		{
			xr = m_shadingRes - 1 - y;
			yr = x;
		}
		else if (r == 1)
		{
			xr = y;
			yr = m_shadingRes - 1 - x;
		}
		else if (r == 2 || r == -2)
		{
			xr = m_shadingRes - 1 - x;
			yr = m_shadingRes - 1 - y;
		}

		x = xr;
		y = yr;
	}

}
