/*
* Copyright (c) 2016, Intel Corporation
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* - Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
* - Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
* - Neither the name of Intel Corporation nor the names of its contributors
*   may be used to endorse or promote products derived from this software
*   without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/

#pragma once

#include <vector>
#include <algorithm>
#include "../common/math/vec2.h"
#include "../common/math/vec3.h"
#include "../common/math/vec3fa.h"

#include "mipcache.h"

namespace embree
{
	/* This struct holds the per sample attributes */
	struct FatSample
	{
		Vec3fa	color;				// rgba 
		Vec3f	normal;				// n.xyz
		int		numSamples;			// Number of samples, used for accumulation

		void accumulate(const FatSample &s)
		{
			float rcp = 1.0f / (float)(numSamples + s.numSamples);
			color = (color*(float)numSamples + s.color*(float)s.numSamples)*rcp;
			normal = (normal*(float)numSamples + s.normal*(float)s.numSamples)*rcp;
			numSamples += s.numSamples;
		}
	};
	
	class Renderer; // Forward declaration

	// Each object is assigned a PatchTexture, which consists of two mip hieararchies for 
	// shading samples and filtered texels. 
	// Note that the connectivity information (face neighbors and edge neighbors) are 
	// stored explicitly, to allow for filtering across multiple faces, similar to Ptex.
	// The main callback from the rendering code is "GetFilteredValue" which lazily 
	// shade, cache and filter within a filter footprint, and returns the filtered value.
	class PatchTexture
	{
	public:

		PatchTexture(const int geomID, const int numFaces, const int shadingRes, 
					 const std::vector<int>& face_neighbors, const std::vector<int>& edge_neighbors, 
					 const std::vector<int>& edges_per_face, const Renderer* pRenderer);

		virtual ~PatchTexture();

		// This is the main callback from the rendering code.
		Vec3fa GetFilteredValue(const int filterRadius, const float mipBias, int primID, float u, float v, const float time, const Vec2f &dUVdX, const Vec2f &dUVdY);

		inline int GetShadingRes() const { return m_shadingRes;	}
		inline int GetNumFaces() const { return m_numFaces; }

		inline void Clear() {
			m_shadingCache->ClearFast();
			m_filterCache->ClearFast();
		}

		inline void ClearFull() {
			m_shadingCache->ClearFull();
			m_filterCache->ClearFull();
		}

		void SetAccumulate(bool val) { m_accumulate = val; }
		
	protected:

		inline bool IsShaded(const int faceID) const {
			return m_shadingCache->GetShade(faceID);
		}
		inline void SetShaded(const int faceID)
		{
			m_shadingCache->SetShade(faceID, true);
		}

		int GetNumMIPLevels() const { return m_numMipLevels; }

		__forceinline float ComputeMIPf(Vec2f &ddx, Vec2f &ddy, float lodBias = 0.0f)
		{
			return min(float(m_numMipLevels - 1), lodBias + 0.5f + (float)max(0.0, log2(min(sqrt(ddx.x*ddx.x + ddx.y*ddx.y), sqrt(ddy.x*ddy.x + ddy.y*ddy.y)))));
		}

		unsigned int ComputeMIP(Vec2f &ddx, Vec2f &ddy, float lodBias = 0.0f)
		{
			return (int)ComputeMIPf(ddx, ddy, lodBias);
		}

		bool GetShadedMIP(int faceID, int mipLevel) const
		{
			return m_shadingCache->GetMipShade(faceID, mipLevel);
		}

		struct FilterWindow {
			int us, vs;
			int ue, ve;
			int iu, iv;
			FatSample *filterTexels;
		};

		Vec3fa FilterTexel(int faceID, const float u, const float v, float mipLevelf, float time, float filterRadius);

		// Finds a sample given a certain (u,v). This may return samples outside the current face, and 
		// relies on patch connectivity information
		void FindSample(unsigned int faceID, int u, int v, unsigned int &outFaceID, int &outU, int &outV, unsigned int depth = 0) const;

		bool GetShadedSampleAsync(int faceID, int x, int y, int mipLevel, float time, FatSample &s);

		inline FatSample &GetSampleMIP(int faceID, int x, int y, int mipLevel)
		{
			return m_shadingCache->GetTexelElement(faceID, x, y, mipLevel).m_data;
		}

		// A simple example filter kernel
		Vec3fa BoxFilterWithNormalThreshold(FilterWindow &window, const FatSample &centerTexel);

		typedef MipCache<FatSample> ShadingCache;
		typedef MipCache<Vec3f>		FilteringCache;

		// Members
		bool								m_accumulate;

		int									m_geomID;
		const Renderer*						m_pRenderer;
		unsigned int						m_numFaces;
		unsigned int						m_shadingRes;
		unsigned int						m_numMipLevels;

		FilteringCache						*m_filterCache;
		ShadingCache						*m_shadingCache;

		int* m_face_neighbors;
		int* m_edge_neighbors;
		int* m_edges_per_face;

		void Rotate(const int r, int &x, int &y) const;

		int GetEdgeNeighbor(int faceID, int curr_edge) const
		{
			return faceID == -1 ? -1 : m_edge_neighbors[4 * faceID + curr_edge];
		}

		int GetFaceNeighbor(int faceID, int edge) const
		{
			return faceID == -1 ? -1 : m_face_neighbors[4 * faceID + edge];
		}

		int GetRotation(int faceID, int e_in) const
		{
			int e = GetEdgeNeighbor(faceID, e_in);

			int r = 0;
			if (m_edges_per_face[faceID] == 4)
				r = (e_in - e + 2) % 4;

			return r;
		}

		/* Compute valence for a face vertex*/
		int ComputeValence(int faceID, int vertex) const;
	};
}
