// ======================================================================== //
// Copyright 2009-2016 Intel Corporation                                    //
//                                                                          //
// Licensed under the Apache License, Version 2.0 (the "License");          //
// you may not use this file except in compliance with the License.         //
// You may obtain a copy of the License at                                  //
//                                                                          //
//     http://www.apache.org/licenses/LICENSE-2.0                           //
//                                                                          //
// Unless required by applicable law or agreed to in writing, software      //
// distributed under the License is distributed on an "AS IS" BASIS,        //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. //
// See the License for the specific language governing permissions and      //
// limitations under the License.                                           //
//																			//
// Modified by Jacob Munkberg												//
// ======================================================================== //

#include "renderer.h"
#include "defines.h"

#include "../support/scenegraph/scene.h"
#include "../support/scenegraph/ray.h"
#include "sceneconversion.h"

static const float g_tnear_ao = 0.01f;
static const float g_tfar_ao  = 1.0f;
static int g_keyframeID = 0;

namespace embree
{
	Renderer::Renderer()
		: m_accu_vx(0.0f)
		, m_accu_vy(0.0f)
		, m_accu_vz(0.0f)
		, m_accu_p(0.0f)
		, m_embree_device(nullptr)
		, m_embree_scene(nullptr)
		, m_width(0)
		, m_height(0)
		, m_pixels(nullptr)
		, m_accu(nullptr)
		, m_scene(nullptr)
		, m_use_smooth_normals(false)
		, m_accu_count(0)
	{}

	/* initialize renderer */
	void Renderer::init(const char* cfg, const bool smooth_normals)
	{
		m_use_smooth_normals = smooth_normals;
		/* initialize last seen camera */
		m_accu_vx = Vec3fa(0.0f);
		m_accu_vy = Vec3fa(0.0f);
		m_accu_vz = Vec3fa(0.0f);
		m_accu_p  = Vec3fa(0.0f);

		/* create new Embree device */
		m_embree_device = rtcNewDevice(cfg);
		SceneConversion::Error_handler(rtcDeviceGetError(m_embree_device));

		/* set error handler */
		rtcDeviceSetErrorFunction(m_embree_device, SceneConversion::Error_handler);
	}

	/* keypressed event */
	void Renderer::key_pressed(int key) {}

	/* resize framebuffer */
	void Renderer::resize(int width, int height) 
	{
		if (width == m_width && height == m_height)
			return;

		if (m_pixels) alignedFree(m_pixels);
		m_width = width;
		m_height = height;
		m_pixels = (int*)alignedMalloc(m_width*m_height*sizeof(int), 64);

		/* create accumulator */
		if (m_accu) alignedFree(m_accu);
		m_accu = (Vec3fa*)alignedMalloc(width*height*sizeof(Vec3fa));
		memset(m_accu, 0, width*height*sizeof(Vec3fa));
	}

	Vec3fa Renderer::ComputeAO(const Vec3fa& P, const Vec3fa& N, RandomSampler& sampler, const float time) const
	{
		float intensity = 0;
		for (int i = 0; i < m_render_param.ao; i++)
		{
			Vec2f uv = sampler.get2D();
			Vec3f dir = cosineSampleHemisphere(uv.x, uv.y, N);

			RTCRay shadow;
			shadow.org = P;
			shadow.dir = dir;
			shadow.tnear = g_tnear_ao;
			shadow.tfar = g_tfar_ao;
			shadow.geomID = RTC_INVALID_GEOMETRY_ID;
			shadow.primID = RTC_INVALID_GEOMETRY_ID;
			shadow.mask = -1;
			shadow.time = time;

			rtcOccluded(m_embree_scene, shadow);

			/* add light contribution */
			if (shadow.geomID == RTC_INVALID_GEOMETRY_ID)
				intensity += 1.0f;
		}
		intensity /= m_render_param.ao;
		return Vec3fa(intensity, intensity, intensity, 1.0f);
	}

	void Renderer::texelCallback(unsigned int geomID, unsigned int primID, int x, int y, int res, float time, FatSample &out) const
	{
		float frac = 1.0f / res;
		float u = (x + 0.5f)*frac;
		float v = (y + 0.5f)*frac;

		// Find position and normal on patch
		Vec3fa Ng(0);
		Vec3fa P(0);

		Vec3fa dPdu, dPdv;
		rtcInterpolate(m_embree_scene, geomID, primID, u, v, RTC_VERTEX_BUFFER, &P.x, &dPdu.x, &dPdv.x, 3);
		Ng = normalize(cross(dPdv, dPdu));

		Vec3fa col = Vec3fa(1, 1, 1);

		Vec3fa Pd = P;
		Vec3fa Ndisp = Ng;

		Vec3fa intensity = Vec3fa(1.0f);
		RandomSampler sampler(x, y, rand());
	
		intensity = ComputeAO(Pd, Ndisp, sampler, time);
		out.color = col*intensity;
		out.normal = Ndisp;
		out.numSamples = m_render_param.ao;
	}

	// Setup patch textures
	void Renderer::SetupPatchTextures()
	{
		// Scene-specific haxxx for now
		const int shading_res = PATCH_SHADING_RES;

		for (auto& geometry : m_scene->geometries)
		{
			if (geometry->type == Geometry::SUBDIV_MESH)
			{
				SubdivMesh* mesh = (SubdivMesh*)geometry.ptr;
				m_patch_textures[mesh->geomID] = new PatchTexture(mesh->geomID, mesh->numFaces, shading_res, 
																  mesh->face_neighbors, mesh->edge_neighbors, 
																  mesh->verticesPerFace, this);
			}
			else if (geometry->type == Geometry::QUAD_MESH)
			{
				assert(0); // TODO: Implement!
			}
		}
	}

	/* pick event */
	bool Renderer::pick(const float x, const float y, const Vec3fa& vx, const Vec3fa& vy, const Vec3fa& vz, const Vec3fa& p, Vec3fa& hitPos) 
	{
		/* initialize ray */
		RTCRay ray;
		ray.org = p;
		ray.dir = normalize(x*vx + y*vy + vz);
		ray.tnear = 0.0f;
		ray.tfar = inf;
		ray.geomID = RTC_INVALID_GEOMETRY_ID;
		ray.primID = RTC_INVALID_GEOMETRY_ID;
		ray.mask = -1;
		ray.time = 0.0f;

		/* intersect ray with scene */
		rtcIntersect(m_embree_scene, ray);
		PRINT2(x, y);
		PRINT(ray.geomID);
		PRINT(ray.primID);

		/* shade pixel */
		if (ray.geomID == RTC_INVALID_GEOMETRY_ID) {
			hitPos = Vec3fa(0.0f, 0.0f, 0.0f);
			return false;
		}
		else {
			hitPos = ray.org + ray.tfar*ray.dir;
			return true;
		}
	}

	// Based on pbrt's ray differential computation
	void Renderer::ComputeUVDerivatives(const RTCRay &ray, const Vec3fa &rxdir, const Vec3fa &rydir, const Vec3fa &p, const Vec3fa &dpdu, const Vec3fa &dpdv, Vec2f &ddx, Vec2f &ddy)
	{
		Vec3fa n = -normalize(cross(dpdu, dpdv));
		float d = -dot(n, p);

		float tx = -(dot(n, ray.org) + d) / dot(n, rxdir);
		Vec3fa px = rxdir*tx + ray.org;

		float ty = -(dot(n, ray.org) + d) / dot(n, rydir);
		Vec3fa py = rydir*ty + ray.org;

		Vec3fa dpdx = px - p;
		Vec3fa dpdy = py - p;

		float A[2][2], Bx[2], By[2];
		int axes[2];

		// Select good subset of equations
		if (fabs(n.x) > fabs(n.y) && fabs(n.x) > fabs(n.z)) {
			axes[0] = 1; axes[1] = 2;
		}
		else if (fabs(n.y) > fabs(n.z)) {
			axes[0] = 0; axes[1] = 2;
		}
		else {
			axes[0] = 0; axes[1] = 1;
		}

		A[0][0] = dpdu[axes[0]];
		A[0][1] = dpdv[axes[0]];
		A[1][0] = dpdu[axes[1]];
		A[1][1] = dpdv[axes[1]];

		Bx[0] = dpdx[axes[0]];
		Bx[1] = dpdx[axes[1]];
		By[0] = dpdy[axes[0]];
		By[1] = dpdy[axes[1]];

		float dudx, dvdx, dudy, dvdy;
		if (!SolveLinearSystem2x2(A, Bx, dudx, dvdx)) {
			dudx = 0; dvdx = 0;
		}
		if (!SolveLinearSystem2x2(A, By, dudy, dvdy)) {
			dudy = 0; dvdy = 0;
		}

		ddx.x = dudx;
		ddx.y = dvdx;

		ddy.x = dudy;
		ddy.y = dvdy;
	}

	/* AO tracer */
	Vec3fa Renderer::RenderPath(float x, float y, RandomSampler& sampler, const Vec3fa& vx, const Vec3fa& vy, const Vec3fa& vz, const Vec3fa& p)
	{
		/* radiance accumulator and weight */
		Vec3fa L = Vec3fa(0.0f);
		float time = 0.0f;

		/* initialize primary ray */
		RTCRay ray(p, RayDir(x, y, vx, vy, vz), 0.0f, inf, time);

		/* intersect primary ray with scene */
		rtcIntersect(m_embree_scene, ray);

		if (ray.geomID == RTC_INVALID_GEOMETRY_ID)
		{
			L = L + Vec3fa(1, 1, 1, 1); // white background
		}
		else {
			Vec3fa Ns = normalize(ray.Ng);

			// Ray differentials...
			Vec3fa rxdir = RayDir(x+1.0f, y, vx, vy, vz);
			Vec3fa rydir = RayDir(x, y+1.0f, vx, vy, vz);

			Vec3fa pos, dpdu, dpdv;
			Vec2f ddx(0), ddy(0);
			rtcInterpolate(m_embree_scene, ray.geomID, ray.primID, ray.u, ray.v, RTC_VERTEX_BUFFER0, &pos.x, &dpdu.x, &dpdv.x, 3);
			ComputeUVDerivatives(ray, rxdir, rydir, pos, dpdu, dpdv, ddx, ddy);
			Ns = normalize(cross(dpdv, dpdu));

			/* Material */
			DifferentialGeometry dg(ray, Ns);
			int materialID = SceneConversion::PostIntersect(ray, m_scene, dg);
			int numMaterials = m_scene->materials.size();
			assert(materialID < numMaterials);
			Material* mat = m_scene->materials[materialID];
			ShaderContext context; // Thread-specific shader context (to avoid texture sampling for each sample)
			const Vec3fa wo = neg(ray.dir);
			mat->preprocess(wo, dg, context);
			Vec3fa c = mat->albedo(context);	

			if (m_render_param.algo == AlgoCache)
			{
				const int filterRadius = m_render_param.filterRadius;
				const float mipBias = m_render_param.mipBias;
				L = c*m_patch_textures[ray.geomID]->GetFilteredValue(filterRadius, mipBias, ray.primID, ray.u, ray.v, time, ddx, ddy);
			} 
			else {
				L = c*ComputeAO(dg.P, dg.Ns, sampler, time);
			}
		}

		return L;
	}

	/* task that renders a single screen tile */
	Vec3fa Renderer::RenderPixel(float x, float y, const Vec3fa& vx, const Vec3fa& vy, const Vec3fa& vz, const Vec3fa& p)
	{
		Vec3fa L = Vec3fa(0.0f);
		RandomSampler sampler(x, y, rand());

		for (int i = 0; i<m_render_param.spp; i++)
		{
			sampler.init(x, y, m_accu_count*m_render_param.spp + i);
			//sampler.init(x, y, rand());

			/* calculate pixel color */
			float fx = x + sampler.get1D();
			float fy = y + sampler.get1D();
			L = L + RenderPath(fx, fy, sampler, vx, vy, vz, p);
		}
		L = L*(1.0f / m_render_param.spp);
		return L;
	}

	/* task that renders a single screen tile */
	void Renderer::RenderTile(int taskIndex, int* pixels,
		const int width,
		const int height,
		const float time,
		const Vec3fa& vx,
		const Vec3fa& vy,
		const Vec3fa& vz,
		const Vec3fa& p,
		const int numTilesX,
		const int numTilesY)
	{
		const int tileY = taskIndex / numTilesX;
		const int tileX = taskIndex - tileY * numTilesX;
		const int x0 = tileX * TILE_SIZE_X;
		const int x1 = min(x0 + TILE_SIZE_X, width);
		const int y0 = tileY * TILE_SIZE_Y;
		const int y1 = min(y0 + TILE_SIZE_Y, height);

		for (int y = y0; y<y1; y++) for (int x = x0; x<x1; x++)
		{
			/* calculate pixel color */
			Vec3fa color = RenderPixel(x, y, vx, vy, vz, p);

			/* write color to framebuffer */
			Vec3fa accu_color = m_accu[y*width + x] + Vec3fa(color.x, color.y, color.z, 1.0f); m_accu[y*width + x] = accu_color;
			float f = rcp(max(0.001f, accu_color.w));
			unsigned int r = (unsigned int)(255.0f * clamp(accu_color.x*f, 0.0f, 1.0f));
			unsigned int g = (unsigned int)(255.0f * clamp(accu_color.y*f, 0.0f, 1.0f));
			unsigned int b = (unsigned int)(255.0f * clamp(accu_color.z*f, 0.0f, 1.0f));
			pixels[y*width + x] = (b << 16) + (g << 8) + r;
		}
	} 
	  /* Launch all renderTile tasks */
	void Renderer::RenderTiles(int numTiles,
		int* pixels, const int width, const int height, const float time,
		const Vec3fa& vx, const Vec3fa& vy, const Vec3fa& vz, const Vec3fa& p, const int numTilesX, const int numTilesY)
	{
#pragma omp parallel for
		for (int i = 0; i < numTiles; ++i)
		{ 
			RenderTile(i, pixels, width, height, time, vx, vy, vz, p, numTilesX, numTilesY);
		}
	}

	/* End render functions */

	/* render frame and map framebuffer */
	void Renderer::render(const float time, const Vec3fa& vx, const Vec3fa& vy, const Vec3fa& vz, const Vec3fa& p, const RenderParam& rp)
	{
		/* reset accumulator */
		bool camera_changed = false;
		camera_changed |= ne(m_accu_vx, vx); m_accu_vx = vx;
		camera_changed |= ne(m_accu_vy, vy); m_accu_vy = vy;
		camera_changed |= ne(m_accu_vz, vz); m_accu_vz = vz;
		camera_changed |= ne(m_accu_p, p);   m_accu_p = p;
		camera_changed |= !(m_render_param == rp); m_render_param = rp;
		camera_changed |= !m_render_param.accumulate;

		Vec3fa cam_org = Vec3fa(p.x, p.y, p.z);

		/* create scene */
		if (m_embree_scene == nullptr)
		{
			m_embree_scene = SceneConversion::ConvertScene(m_scene, m_embree_device);
			SetupPatchTextures();
		}

		if (m_render_param.algo == AlgoCache)
		{
			for (auto &id : m_patch_textures)
			{
				id.second->SetAccumulate(false);
				id.second->Clear();
			}
		}

#if  FIXED_SAMPLING == 0
		m_accu_count++;
#endif

		if (camera_changed) {
			m_accu_count = 0;
			memset(m_accu, 0, m_width*m_height*sizeof(Vec3fa));
		}

		/* render image */
		const int numTilesX = (m_width + TILE_SIZE_X - 1) / TILE_SIZE_X;
		const int numTilesY = (m_height + TILE_SIZE_Y - 1) / TILE_SIZE_Y;
		RenderTiles(numTilesX*numTilesY, m_pixels, m_width, m_height, time, vx, vy, vz, p, numTilesX, numTilesY);
	}

	/* cleanup renderer */
	void Renderer::cleanup() 
	{
		if (m_pixels) alignedFree(m_pixels);
		if (m_accu) alignedFree(m_accu);

		rtcDeleteScene(m_embree_scene);
		rtcDeleteDevice(m_embree_device);
	}
};