// ======================================================================== //
// Copyright 2009-2016 Intel Corporation                                    //
//                                                                          //
// Licensed under the Apache License, Version 2.0 (the "License");          //
// you may not use this file except in compliance with the License.         //
// You may obtain a copy of the License at                                  //
//                                                                          //
//     http://www.apache.org/licenses/LICENSE-2.0                           //
//                                                                          //
// Unless required by applicable law or agreed to in writing, software      //
// distributed under the License is distributed on an "AS IS" BASIS,        //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. //
// See the License for the specific language governing permissions and      //
// limitations under the License.                                           //
//																			//
// Modified by Jacob Munkberg												//
// ======================================================================== //

#pragma once

#include <embree2/rtcore.h>

#include "../common/math/vec3.h"
#include "../support/scenegraph/differentialgeometry.h"
#include "patchtexture.h"
#include "renderparams.h"
#include "../support/scenegraph/sampler.h"
#include "../support/scenegraph/materials.h"

namespace embree
{
	// Forward declarations
	struct Scene;
	struct RenderParam;

	class Renderer {

	public:
		Renderer();

		/* initialize renderer */
		void init(const char* cfg, const bool smooth_normals=true);

		/* keypressed event */
		void key_pressed(int key);

		/* resize framebuffer */
		void resize(int width, int height);

		/* set scene to use */
		void set_scene(Scene* in) { m_scene = in; }

		/* pick event */
		bool pick(const float x, const float y, const Vec3fa& vx, const Vec3fa& vy, const Vec3fa& vz, const Vec3fa& p, Vec3fa& hitPos);

		/* render frame and map framebuffer */
		void render(const float time, const Vec3fa& vx, const Vec3fa& vy, const Vec3fa& vz, const Vec3fa& p, const RenderParam& rp);

		/* texel space callback function for lazy shader evaluation */
		void texelCallback(unsigned int geomID, unsigned int primID, int x, int y, int res, float time, FatSample &out) const;

		/* map framebuffer */
		int* map() { return m_pixels;  }

		/* unmap framebuffer */
		void unmap() {}

		/* cleanup renderer */
		void cleanup();

	private:

		/* computes ambient occlusion at an intersection by tracing rays in a hemisphere */
		Vec3fa ComputeAO(const Vec3fa& P, const Vec3fa& N, RandomSampler& sampler, const float time) const;

		/* trace one path */
		Vec3fa RenderPath(float x, float y, RandomSampler& sampler, const Vec3fa& vx, const Vec3fa& vy, const Vec3fa& vz, const Vec3fa& p);

		/* task that renders a single screen tile */
		Vec3fa RenderPixel(float x, float y, const Vec3fa& vx, const Vec3fa& vy, const Vec3fa& vz, const Vec3fa& p);

		/* task that renders a single screen tile */
		void RenderTile(int taskIndex, int* pixels,
			const int width, const int height, const float time,
			const Vec3fa& vx, const Vec3fa& vy, const Vec3fa& vz,
			const Vec3fa& p, const int numTilesX, const int numTilesY);

		/* launch all renderTile tasks using tbb*/
		void RenderTiles(int numTiles,
			int* pixels, const int width, const int height, const float time,
			const Vec3fa& vx, const Vec3fa& vy, const Vec3fa& vz, const Vec3fa& p, 
			const int numTilesX, const int numTilesY);

		/* setup and allocate memory for texture space shading structs */
		void SetupPatchTextures();

		/* Helper function for linear system, taken from pbrt */
		inline bool SolveLinearSystem2x2(const float A[2][2],
			const float B[2], float& x0, float& x1) {

			float det = A[0][0] * A[1][1] - A[0][1] * A[1][0];
			if (fabs(det) < 1e-10f) {
				return false;
			}
			x0 = (A[1][1] * B[0] - A[0][1] * B[1]) / det;
			x1 = (A[0][0] * B[1] - A[1][0] * B[0]) / det;

			if (isnan(x0) || isnan(x1)) {
				return false;
			}
			return true;
		}

		inline Vec3fa RayDir(float x, float y, const Vec3fa &vx, const Vec3fa &vy, const Vec3fa &vz)
		{
			return normalize(x*vx + y*vy + vz);
		}

		void ComputeUVDerivatives(const RTCRay &ray, const Vec3fa &rxdir, const Vec3fa &rydir, const Vec3fa &p,
			                      const Vec3fa &dpdu, const Vec3fa &dpdv, Vec2f &ddx, Vec2f &ddy);

	private:
		static const int TILE_SIZE_X = 4;
		static const int TILE_SIZE_Y = 4;

		Vec3fa m_accu_vx;
		Vec3fa m_accu_vy;
		Vec3fa m_accu_vz;
		Vec3fa m_accu_p;

		int m_width; 		// Framebuffer
		int m_height;
		int* m_pixels;
		Vec3fa* m_accu;
		int m_accu_count;

		Scene* m_scene;			// Scene
		bool m_use_smooth_normals;

		RTCDevice m_embree_device; 		// Embree structs
		RTCScene  m_embree_scene;

		std::map<int, PatchTexture*> m_patch_textures; // PatchTextures (holds shading values)
		RenderParam  m_render_param;				   // Render parameters
	};
}