// ======================================================================== //
// Copyright 2009-2016 Intel Corporation                                    //
//                                                                          //
// Licensed under the Apache License, Version 2.0 (the "License");          //
// you may not use this file except in compliance with the License.         //
// You may obtain a copy of the License at                                  //
//                                                                          //
//     http://www.apache.org/licenses/LICENSE-2.0                           //
//                                                                          //
// Unless required by applicable law or agreed to in writing, software      //
// distributed under the License is distributed on an "AS IS" BASIS,        //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. //
// See the License for the specific language governing permissions and      //
// limitations under the License.                                           //
//																			//
// Modified by Jacob Munkberg												//
// ======================================================================== //

#include "sceneconversion.h"
#include "../support/scenegraph/scene.h"

#define FIXED_EDGE_TESSELLATION_VALUE 32.0f
#define MAX_EDGE_LEVEL 128.0f
#define MIN_EDGE_LEVEL   4.0f
#define LEVEL_FACTOR     64.0f

namespace embree
{
	/* error reporting function */
	void SceneConversion::Error_handler(const RTCError code, const char* str)
	{
		if (code == RTC_NO_ERROR)
			return;

		printf("Embree: ");
		switch (code) {
		case RTC_UNKNOWN_ERROR: printf("RTC_UNKNOWN_ERROR"); break;
		case RTC_INVALID_ARGUMENT: printf("RTC_INVALID_ARGUMENT"); break;
		case RTC_INVALID_OPERATION: printf("RTC_INVALID_OPERATION"); break;
		case RTC_OUT_OF_MEMORY: printf("RTC_OUT_OF_MEMORY"); break;
		case RTC_UNSUPPORTED_CPU: printf("RTC_UNSUPPORTED_CPU"); break;
		case RTC_CANCELLED: printf("RTC_CANCELLED"); break;
		default: printf("invalid error code"); break;
		}
		if (str) {
			printf(" (");
			while (*str) putchar(*str++);
			printf(")\n");
		}
		exit(1);
	} // error handler

	unsigned int convertTriangleMesh(TriangleMesh* mesh, RTCScene scene_out)
	{
		unsigned int geomID = rtcNewTriangleMesh(scene_out, RTC_GEOMETRY_STATIC, mesh->triangles.size(), mesh->v.size(), mesh->v2.size() > 0 ? 2 : 1);
		rtcSetBuffer(scene_out, geomID, RTC_VERTEX_BUFFER, mesh->v.data(), 0, sizeof(Vec3fa));
		if (mesh->v2.size() > 0) rtcSetBuffer(scene_out, geomID, RTC_VERTEX_BUFFER1, mesh->v2.data(), 0, sizeof(Vec3fa));
		rtcSetBuffer(scene_out, geomID, RTC_INDEX_BUFFER, mesh->triangles.data(), 0, sizeof(Scene::Triangle));
		mesh->geomID = geomID;
		return geomID;
	}

	unsigned int convertQuadMesh(QuadMesh* mesh, RTCScene scene_out)
	{
		unsigned int geomID = rtcNewQuadMesh(scene_out, RTC_GEOMETRY_STATIC, mesh->quads.size(), mesh->v.size(), mesh->v2.size() > 0 ? 2 : 1);
		rtcSetBuffer(scene_out, geomID, RTC_VERTEX_BUFFER, mesh->v.data(), 0, sizeof(Vec3fa));
		if (mesh->v2.size() > 0) rtcSetBuffer(scene_out, geomID, RTC_VERTEX_BUFFER1, mesh->v2.data(), 0, sizeof(Vec3fa));
		rtcSetBuffer(scene_out, geomID, RTC_INDEX_BUFFER, mesh->quads.data(), 0, sizeof(Scene::Quad));
		mesh->geomID = geomID;
		return geomID;
	}

	unsigned int convertSubdivMesh(SubdivMesh* mesh, RTCScene scene_out)
	{
		unsigned int geomID = rtcNewSubdivisionMesh(scene_out, RTC_GEOMETRY_DYNAMIC, mesh->numFaces, mesh->numEdges, mesh->numVertices,
			mesh->numEdgeCreases, mesh->numVertexCreases, mesh->numHoles);
		mesh->geomID = geomID;
		for (size_t i = 0; i<mesh->numEdges; i++) mesh->subdivlevel[i] = FIXED_EDGE_TESSELLATION_VALUE;
		rtcSetBuffer(scene_out, geomID, RTC_VERTEX_BUFFER, mesh->positions.data(), 0, sizeof(Vec3fa));
		rtcSetBuffer(scene_out, geomID, RTC_LEVEL_BUFFER, mesh->subdivlevel, 0, sizeof(float));
		rtcSetBuffer(scene_out, geomID, RTC_INDEX_BUFFER, mesh->position_indices.data(), 0, sizeof(unsigned int));
		rtcSetBuffer(scene_out, geomID, RTC_FACE_BUFFER, mesh->verticesPerFace.data(), 0, sizeof(unsigned int));
		rtcSetBuffer(scene_out, geomID, RTC_HOLE_BUFFER, mesh->holes.data(), 0, sizeof(unsigned int));
		rtcSetBuffer(scene_out, geomID, RTC_EDGE_CREASE_INDEX_BUFFER, mesh->edge_creases.data(), 0, 2 * sizeof(unsigned int));
		rtcSetBuffer(scene_out, geomID, RTC_EDGE_CREASE_WEIGHT_BUFFER, mesh->edge_crease_weights.data(), 0, sizeof(float));
		rtcSetBuffer(scene_out, geomID, RTC_VERTEX_CREASE_INDEX_BUFFER, mesh->vertex_creases.data(), 0, sizeof(unsigned int));
		rtcSetBuffer(scene_out, geomID, RTC_VERTEX_CREASE_WEIGHT_BUFFER, mesh->vertex_crease_weights.data(), 0, sizeof(float));
		return geomID;
	}

	void convertGroup(Group* group, RTCScene scene_out)
	{
		for (size_t i = 0; i<group->size(); i++)
		{
			Ref<Geometry> geometry = group->at(i);
			if (geometry->type == Geometry::SUBDIV_MESH)
			{
				convertSubdivMesh((SubdivMesh*)geometry.ptr, scene_out);
			}
			else if (geometry->type == Geometry::TRIANGLE_MESH)
				convertTriangleMesh((TriangleMesh*)geometry.ptr, scene_out);
			else if (geometry->type == Geometry::QUAD_MESH)
				convertQuadMesh((QuadMesh*)geometry.ptr, scene_out);
			else if (geometry->type == Geometry::LINE_SEGMENTS)
				assert(0);
			else if (geometry->type == Geometry::HAIR_SET)
				assert(0);
			else
				assert(false);
		}
	}

	RTCScene SceneConversion::ConvertScene(Scene* scene_in, RTCDevice device)
	{
		size_t numGeometries = scene_in->geometries.size();

		bool subdiv_mode = false;
		for (size_t i = 0; i < numGeometries; i++)
		{
			Ref<Geometry> geometry = scene_in->geometries[i];
			if (geometry->type == Geometry::SUBDIV_MESH) {
				subdiv_mode = true; break;
			}
		}

		/* create scene */
		int scene_flags = RTC_SCENE_STATIC | RTC_SCENE_INCOHERENT;
		int scene_aflags = RTC_INTERSECT1;

		if (subdiv_mode)
			scene_flags = RTC_SCENE_DYNAMIC | RTC_SCENE_INCOHERENT | RTC_SCENE_ROBUST;

		scene_aflags |= RTC_INTERPOLATE;

		RTCScene scene_out = rtcDeviceNewScene(device, (RTCSceneFlags)scene_flags, (RTCAlgorithmFlags)scene_aflags);

		/* no instancing */
		for (size_t i = 0; i < numGeometries; i++)
		{
			Ref<Geometry> geometry = scene_in->geometries[i];
			if (geometry->type == Geometry::SUBDIV_MESH) {
				SubdivMesh* pSubd = (SubdivMesh*)geometry.ptr;
				unsigned int geomID = convertSubdivMesh(pSubd, scene_out);
				assert(geomID == i);
			}
			else if (geometry->type == Geometry::TRIANGLE_MESH) {
				unsigned int geomID = convertTriangleMesh((TriangleMesh*)geometry.ptr, scene_out);
				assert(geomID == i);
			}
			else if (geometry->type == Geometry::QUAD_MESH) {
				unsigned int geomID = convertQuadMesh((QuadMesh*)geometry.ptr, scene_out);
				assert(geomID == i);
			}
			else if (geometry->type == Geometry::LINE_SEGMENTS) {
				assert(0);
			}
			else if (geometry->type == Geometry::HAIR_SET) {
				assert(0);
			}
			else
				assert(false);
		}

		/* commit changes to scene */
		rtcCommit(scene_out);
		return scene_out;
	}

	Vec2f getTextureCoordinatesSubdivMesh(void* _mesh, const unsigned int primID, const float u, const float v)
	{
		SubdivMesh *mesh = (SubdivMesh*)_mesh;
		Vec2f st;
		st.x = u;
		st.y = v;
		if (mesh && mesh->texcoord_indices.size() > 0 && mesh->texcoords.size() > 0)
		{
			assert(primID < mesh->numFaces);
			const size_t face_offset = mesh->face_offsets[primID];
			if (mesh->verticesPerFace[primID] == 3)
			{
				const size_t t0 = mesh->texcoord_indices[face_offset + 0];
				const size_t t1 = mesh->texcoord_indices[face_offset + 1];
				const size_t t2 = mesh->texcoord_indices[face_offset + 2];
				const Vec2f &txt0 = mesh->texcoords[t0];
				const Vec2f &txt1 = mesh->texcoords[t1];
				const Vec2f &txt2 = mesh->texcoords[t2];
				const float w = 1.0f - u - v;
				st = w * txt0 + u * txt1 + v * txt2;
			}
			else if (mesh->verticesPerFace[primID] == 4)
			{
				const size_t t0 = mesh->texcoord_indices[face_offset + 0];
				const size_t t1 = mesh->texcoord_indices[face_offset + 1];
				const size_t t2 = mesh->texcoord_indices[face_offset + 2];
				const size_t t3 = mesh->texcoord_indices[face_offset + 3];
				const Vec2f &txt0 = mesh->texcoords[t0];
				const Vec2f &txt1 = mesh->texcoords[t1];
				const Vec2f &txt2 = mesh->texcoords[t2];
				const Vec2f &txt3 = mesh->texcoords[t3];
				const float u0 = u;
				const float v0 = v;
				const float u1 = 1.0f - u;
				const float v1 = 1.0f - v;
				st = u1*v1 * txt0 + u0*v1* txt1 + u0*v0 * txt2 + u1*v0* txt3;
			}
#if defined(_DEBUG)
			else
				PRINT("not supported");
#endif
		}
		return st;
	}

	void postIntersectGeometry(const RTCRay& ray, DifferentialGeometry& dg, Geometry* geometry, int& materialID)
	{
		if (geometry->type == Geometry::TRIANGLE_MESH)
		{
			TriangleMesh* mesh = (TriangleMesh*)geometry;
			materialID = mesh->triangles[ray.primID].materialID;
			if (mesh->vt.size() > 0) {
				Triangle* tri = &mesh->triangles[ray.primID];
				const Vec2f st0 = Vec2f(mesh->vt[tri->v0]);
				const Vec2f st1 = Vec2f(mesh->vt[tri->v1]);
				const Vec2f st2 = Vec2f(mesh->vt[tri->v2]);
				const float u = ray.u, v = ray.v, w = 1.0f - ray.u - ray.v;
				const Vec2f st = w*st0 + u*st1 + v*st2;
				dg.u = st.x;
				dg.v = st.y;
			}
		}
		else if (geometry->type == Geometry::QUAD_MESH)
		{
			QuadMesh* mesh = (QuadMesh*)geometry;
			materialID = mesh->meshMaterialID;
			if (mesh->vt.size() > 0) {
				Quad* quad = &mesh->quads[ray.primID];
				const Vec2f st0 = Vec2f(mesh->vt[quad->v0]);
				const Vec2f st1 = Vec2f(mesh->vt[quad->v1]);
				const Vec2f st2 = Vec2f(mesh->vt[quad->v2]);
				const Vec2f st3 = Vec2f(mesh->vt[quad->v3]);
				if (ray.u + ray.v < 1.0f) {
					const float u = ray.u, v = ray.v; const float w = 1.0f - u - v;
					const Vec2f st = w*st0 + u*st1 + v*st3;
					dg.u = st.x;
					dg.v = st.y;
				}
				else {
					const float u = 1.0f - ray.u, v = 1.0f - ray.v; const float w = 1.0f - u - v;
					const Vec2f st = w*st2 + u*st3 + v*st1;
					dg.u = st.x;
					dg.v = st.y;
				}
			}
		}
		else if (geometry->type == Geometry::SUBDIV_MESH)
		{
			SubdivMesh* mesh = (SubdivMesh*)geometry;
			materialID = mesh->materialID;
			const Vec2f st = getTextureCoordinatesSubdivMesh(mesh, ray.primID, ray.u, ray.v);
			dg.u = st.x;
			dg.v = st.y;
		}
		else if (geometry->type == Geometry::LINE_SEGMENTS)
		{
			assert(0);
		}
		else if (geometry->type == Geometry::HAIR_SET)
		{
			assert(0);
		}
		else if (geometry->type == Geometry::GROUP) {
			assert(0);
		}
		else
			assert(false);
	}

	int SceneConversion::PostIntersect(const RTCRay& ray, Scene* tut_scene, DifferentialGeometry& dg)
	{
		int materialID = 0;
		unsigned ray_geomID = ray.geomID;
		dg.tnear_eps = 32.0f*1.19209e-07f*max((float)max(abs(dg.P.x), abs(dg.P.y)), max((float)abs(dg.P.z), ray.tfar));
		int geomID = ray_geomID;
		{
			/* get geometry pointers */
			Ref<Geometry> geometry = tut_scene->geometries[geomID];
			postIntersectGeometry(ray, dg, geometry.ptr, materialID);
		}
		return materialID;
	}

	int SceneConversion::GetMaterialID(Scene* tut_scene, const int geomID, const int primID)
	{
		int materialID = 0;
		Ref<Geometry> geometry = tut_scene->geometries[geomID];

		if (geometry->type == Geometry::TRIANGLE_MESH)
		{
			TriangleMesh* mesh = (TriangleMesh*)geometry.ptr;
			materialID = mesh->triangles[primID].materialID;
		}
		else if (geometry->type == Geometry::QUAD_MESH)
		{
			QuadMesh* mesh = (QuadMesh*)geometry.ptr;
			materialID = mesh->meshMaterialID;
		}
		else if (geometry->type == Geometry::SUBDIV_MESH)
		{
			SubdivMesh* mesh = (SubdivMesh*)geometry.ptr;
			materialID = mesh->materialID;
		}
		else {
			assert(false);
		}
		return materialID;
	}
}