// ======================================================================== //
// Copyright 2009-2016 Intel Corporation                                    //
//                                                                          //
// Licensed under the Apache License, Version 2.0 (the "License");          //
// you may not use this file except in compliance with the License.         //
// You may obtain a copy of the License at                                  //
//                                                                          //
//     http://www.apache.org/licenses/LICENSE-2.0                           //
//                                                                          //
// Unless required by applicable law or agreed to in writing, software      //
// distributed under the License is distributed on an "AS IS" BASIS,        //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. //
// See the License for the specific language governing permissions and      //
// limitations under the License.                                           //
//																			//
// Modified by Jacob Munkberg												//
// ======================================================================== //

#include "glutdisplay.h"
#include "../../common/sys/filename.h"
#include "../../common/sys/sysinfo.h"
#include "../../common/lexers/streamfilters.h"
#include "../../common/lexers/parsestream.h"
#include "../../renderer/renderer.h"
#include "../../renderer/renderparams.h"

#include <map>
#include <sstream>
#include <imgui.h>

/* include GLUT for display */
#if defined(__MACOSX__)
#  include <OpenGL/gl.h>
#  include <GLUT/glut.h>
#elif defined(__WIN32__)
#  include <windows.h>
#  include <GL/gl.h>   
#  include <GL/glut.h>
#else
#  include <GL/gl.h>   
#  include <GL/glut.h>
#endif

extern "C" float g_debug = 0.0f;

namespace embree
{
	IMGUI_API void        ImGui_ImplGlfw_NewFrame();

  static const double g_time0 = getSeconds();

  /* camera */
  Camera g_camera;

  /* output settings */
  static size_t g_width = 512;
  static size_t g_height = 512;
  static bool g_display = true;

  /* fullscreen settings */
  static bool g_fullscreen = false;
  static size_t g_window_width = 512;
  static size_t g_window_height = 512;

  /* ID of created window */
  static int g_window = 0;

  Renderer* g_pRenderer = nullptr;
  RenderParam* g_pRenderParam = nullptr;

  /* ImGui */
  static bool g_MousePressed[3] = { false, false, false };
  static GLuint g_FontTexture = 0;
  static int g_cursorX = 0;
  static int g_cursorY = 0;
  static float g_Time = 0.0f;

  static float g_sigma_s = 5.0f;
  static float g_sigma_r = 0.14f;
  /*************************************************************************************************/
  /*                                  Keyboard control                                             */
  /*************************************************************************************************/

  typedef void (* KeyBinding)(unsigned char key, int x, int y);

  static float g_speed = 1.0f;
  static std::map<unsigned char, KeyBinding> keyBindings;

  void mapKeyToFunction(unsigned char key, KeyBinding binding)
  {
    keyBindings[key] = binding;
  }

  void keyboardFunc(unsigned char key, int x, int y)
  {

    if (keyBindings.find(key) != keyBindings.end()) { keyBindings[key](key, x, y);  return; }

    g_pRenderer->key_pressed(key);

    switch (key)
    {
    case 'f' : 
      if (g_fullscreen) {
        g_fullscreen = false;
        glutReshapeWindow(g_window_width,g_window_height);
      } else {
        g_fullscreen = true;
        g_window_width = g_width;
        g_window_height = g_height;
        glutFullScreen(); 
      }
      break;
    case 'c' : {
      std::cout.precision(10);
      std::cout << "-vp " << g_camera.from.x    << " " << g_camera.from.y    << " " << g_camera.from.z    << " " 
                << "-vi " << g_camera.to.x << " " << g_camera.to.y << " " << g_camera.to.z << " " 
                << "-vu " << g_camera.up.x     << " " << g_camera.up.y     << " " << g_camera.up.z     << " " 
                << "-fov " << g_camera.fov << std::endl;
      break;
    }

    case '\033': case 'q': case 'Q':
      g_pRenderer->cleanup();
      glutDestroyWindow(g_window);
#if defined(__MACOSX__)
      exit(1);
#endif
      break;
    }

	ImGuiIO& io = ImGui::GetIO();
	if (key > 0 && key < 0x10000)
		io.AddInputCharacter((unsigned short)key);
  }

  void specialFunc(int key, int, int)
  {
    g_pRenderer->key_pressed(key);

    switch (key) {
    case GLUT_KEY_LEFT      : g_camera.rotate(-0.02f,0.0f); break;
    case GLUT_KEY_RIGHT     : g_camera.rotate(+0.02f,0.0f); break;
    case GLUT_KEY_UP        : g_camera.move(0.0f,0.0f,+g_speed); break;
    case GLUT_KEY_DOWN      : g_camera.move(0.0f,0.0f,-g_speed); break;
    case GLUT_KEY_PAGE_UP   : g_speed *= 1.2f; break;
    case GLUT_KEY_PAGE_DOWN : g_speed /= 1.2f; break;
    }
  }

  /*************************************************************************************************/
  /*                                   Mouse control                                               */
  /*************************************************************************************************/

  static int mouseMode = 0;
  static int clickX = 0, clickY = 0;
  static bool flip14 = false;

  void clickFunc(int button, int state, int x, int y) 
  {
    if (state == GLUT_UP) 
    {
		// ImGui
		if (button >= 0 && button < 3) {
			g_MousePressed[button] = true;
		}

      mouseMode = 0;
      if (button == GLUT_LEFT_BUTTON && glutGetModifiers() == GLUT_ACTIVE_SHIFT) 
      {
        AffineSpace3fa pixel2world = g_camera.pixel2world(g_width,g_height);
        Vec3fa p; bool hit = g_pRenderer->pick(x,y, pixel2world.l.vx, pixel2world.l.vy, pixel2world.l.vz, pixel2world.p, p);

        if (hit) {
          Vec3fa delta = p - g_camera.to;
          Vec3fa right = normalize(pixel2world.l.vx);
          Vec3fa up    = normalize(pixel2world.l.vy);
          g_camera.to = p;
          g_camera.from += dot(delta,right)*right + dot(delta,up)*up;
        }
      }
      else if (button == GLUT_LEFT_BUTTON && glutGetModifiers() == (GLUT_ACTIVE_CTRL | GLUT_ACTIVE_SHIFT)) 
      {
        AffineSpace3fa pixel2world = g_camera.pixel2world(g_width,g_height);
        Vec3fa p; bool hit = g_pRenderer->pick(x,y, pixel2world.l.vx, pixel2world.l.vy, pixel2world.l.vz, pixel2world.p, p);
        if (hit) g_camera.to = p;
      }
    } 
	else 
	{
      clickX = x; clickY = y;
      int modifiers = glutGetModifiers();
      if      (button == GLUT_LEFT_BUTTON && modifiers == GLUT_ACTIVE_SHIFT) mouseMode = 1;
      else if (button == GLUT_MIDDLE_BUTTON) mouseMode = 2;
      else if (button == GLUT_RIGHT_BUTTON ) mouseMode = 3;
      else if (button == GLUT_LEFT_BUTTON && modifiers == GLUT_ACTIVE_CTRL ) mouseMode = 3;
      else if (button == GLUT_LEFT_BUTTON  ) mouseMode = 4;

      if (flip14) {
        if (mouseMode == 4) mouseMode = 1;
        else if (mouseMode == 1) mouseMode = 4;
      }
    }
  }

  void cursorPos(int x, int y)
  {
	  g_cursorX = x;
	  g_cursorY = y;
  }

  void motionFunc(int x, int y)
  {
	  ImGuiIO& io = ImGui::GetIO();
	  if (mouseMode == 1)
	  {
		  g_MousePressed[0] = true;
	  }
	  g_cursorX = x;
	  g_cursorY = y;

	  if (!io.WantCaptureMouse)
	  {
		  float dClickX = float(clickX - x), dClickY = float(clickY - y);
		  clickX = x; clickY = y;

		  switch (mouseMode) {
		  case 1: g_camera.rotateOrbit(-0.005f*dClickX, 0.005f*dClickY); break;
		  case 2: break;
		  case 3: g_camera.dolly(-dClickY); break;
		  case 4: g_camera.rotate(-0.005f*dClickX, 0.005f*dClickY); break;
		  }
	  }
  }
 
  /*************************************************************************************************/
  /*                                   Window control                                              */
  /*************************************************************************************************/

  void displayFunc(void) 
  {
    AffineSpace3fa pixel2world = g_camera.pixel2world(g_width,g_height);

    /* render image */
    double t0 = getSeconds();
	g_pRenderer->render(g_time0 - t0,
		pixel2world.l.vx,
		-pixel2world.l.vy,
		pixel2world.l.vz + g_height*pixel2world.l.vy,
		pixel2world.p, *g_pRenderParam);
    double dt0 = getSeconds()-t0;

    if (g_display) 
    {
      /* draw pixels to screen */
	  int* pixels = g_pRenderer->map();
      glDrawPixels(g_width,g_height,GL_RGBA,GL_UNSIGNED_BYTE,pixels);

      if (g_fullscreen) 
      {
        glMatrixMode( GL_PROJECTION );
        glPushMatrix();
        glLoadIdentity();
        gluOrtho2D( 0, g_width, 0, g_height );
        glMatrixMode( GL_MODELVIEW );
        glPushMatrix();
        glLoadIdentity();

         /* print frame rate */
        std::ostringstream stream;
        stream.setf(std::ios::fixed, std::ios::floatfield);
        stream.precision(2);
        stream << 1.0f/dt0 << " fps";
        std::string str = stream.str();

        glRasterPos2i( g_width-str.size()*12, g_height - 24); 
        for ( int i = 0; i < str.size(); ++i )
          glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_24, str[i]);
        
        glRasterPos2i( 0, 0 ); 
        glPopMatrix();
        glMatrixMode( GL_PROJECTION );
        glPopMatrix();
        glMatrixMode( GL_MODELVIEW );
      }

	  // Draw GUI
#if 1
	  ImGui_ImplGlfw_NewFrame();

	  // Menu
	  ImGui::Combo("Algo", (int*)&g_pRenderParam->algo, "Path\0Cache\0");
	  ImGui::SliderInt("SPP", &g_pRenderParam->spp, 1, 16);
	  ImGui::SliderInt("AO Rays", &g_pRenderParam->ao, 1, 16);

	  if (g_pRenderParam->algo == AlgoCache)
	  {
		  ImGui::SliderFloat("F.size", &g_pRenderParam->filterRadius, 0.0f, 8.0f);
		  ImGui::SliderFloat("MipBias", &g_pRenderParam->mipBias, 0.0f, 4.0f);
	  }
	  ImGui::Checkbox("Accum:", &g_pRenderParam->accumulate);

	  // Rendering
	  ImGui::Render();
#endif
      
      glutSwapBuffers();
      g_pRenderer->unmap();
    }
    double dt1 = getSeconds()-t0;

    /* print frame rate */
    std::ostringstream stream;
    stream.setf(std::ios::fixed, std::ios::floatfield);
    stream.precision(2);
    stream << "render: ";
    stream << 1.0f/dt0 << " fps, ";
    stream << dt0*1000.0f << " ms, ";
    stream << "display: ";
    stream << 1.0f/dt1 << " fps, ";
    stream << dt1*1000.0f << " ms, ";
    stream << g_width << "x" << g_height << " pixels";
    std::cout << stream.str() << std::endl;
  }

  void reshapeFunc(int width, int height) 
  {
	g_pRenderer->resize(width, height);
    glViewport(0, 0, width, height);
    g_width = width; g_height = height;
  }
  
  void idleFunc()
  {
    glutPostRedisplay();
  }

  void enterWindowRunLoop(bool anim)
  {
    if (anim)
      g_camera.enableAnimMode();

    glutMainLoop();
  }

  // ============================== BEGIN IMGUI ==============================

  // This is the main rendering function that you have to implement and provide to ImGui (via setting up 'RenderDrawListsFn' in the ImGuiIO structure)
  // If text or lines are blurry when integrating ImGui in your engine:
  // - in your Render function, try translating your projection matrix by (0.5f,0.5f) or (0.375f,0.375f)
  void ImGui_ImplGlfw_RenderDrawLists(ImDrawData* draw_data)
  {
	  // We are using the OpenGL fixed pipeline to make the example code simpler to read!
	  // Setup render state: alpha-blending enabled, no face culling, no depth testing, scissor enabled, vertex/texcoord/color pointers.
	  GLint last_texture; glGetIntegerv(GL_TEXTURE_BINDING_2D, &last_texture);
	  GLint last_viewport[4]; glGetIntegerv(GL_VIEWPORT, last_viewport);
	  glPushAttrib(GL_ENABLE_BIT | GL_COLOR_BUFFER_BIT | GL_TRANSFORM_BIT);
	  glEnable(GL_BLEND);
	  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	  glDisable(GL_CULL_FACE);
	  glDisable(GL_DEPTH_TEST);
	  glEnable(GL_SCISSOR_TEST);
	  glEnableClientState(GL_VERTEX_ARRAY);
	  glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	  glEnableClientState(GL_COLOR_ARRAY);
	  glEnable(GL_TEXTURE_2D);
	  //glUseProgram(0); // You may want this if using this code in an OpenGL 3+ context

	  // Handle cases of screen coordinates != from framebuffer coordinates (e.g. retina displays)
	  ImGuiIO& io = ImGui::GetIO();
	  float fb_height = io.DisplaySize.y * io.DisplayFramebufferScale.y;
	  draw_data->ScaleClipRects(io.DisplayFramebufferScale);

	  // Setup viewport, orthographic projection matrix
	  glViewport(0, 0, (GLsizei)io.DisplaySize.x, (GLsizei)io.DisplaySize.y);
	  glMatrixMode(GL_PROJECTION);
	  glPushMatrix();
	  glLoadIdentity();
	  glOrtho(0.0f, io.DisplaySize.x, io.DisplaySize.y, 0.0f, -1.0f, +1.0f);
	  glMatrixMode(GL_MODELVIEW);
	  glPushMatrix();
	  glLoadIdentity();

	  // Render command lists
#define OFFSETOF(TYPE, ELEMENT) ((size_t)&(((TYPE *)0)->ELEMENT))
	  for (int n = 0; n < draw_data->CmdListsCount; n++)
	  {
		  const ImDrawList* cmd_list = draw_data->CmdLists[n];
		  const unsigned char* vtx_buffer = (const unsigned char*)&cmd_list->VtxBuffer.front();
		  const ImDrawIdx* idx_buffer = &cmd_list->IdxBuffer.front();
		  glVertexPointer(2, GL_FLOAT, sizeof(ImDrawVert), (void*)(vtx_buffer + OFFSETOF(ImDrawVert, pos)));
		  glTexCoordPointer(2, GL_FLOAT, sizeof(ImDrawVert), (void*)(vtx_buffer + OFFSETOF(ImDrawVert, uv)));
		  glColorPointer(4, GL_UNSIGNED_BYTE, sizeof(ImDrawVert), (void*)(vtx_buffer + OFFSETOF(ImDrawVert, col)));

		  for (int cmd_i = 0; cmd_i < cmd_list->CmdBuffer.size(); cmd_i++)
		  {
			  const ImDrawCmd* pcmd = &cmd_list->CmdBuffer[cmd_i];
			  if (pcmd->UserCallback)
			  {
				  pcmd->UserCallback(cmd_list, pcmd);
			  }
			  else
			  {
				  glBindTexture(GL_TEXTURE_2D, (GLuint)(intptr_t)pcmd->TextureId);
				  glScissor((int)pcmd->ClipRect.x, (int)(fb_height - pcmd->ClipRect.w), (int)(pcmd->ClipRect.z - pcmd->ClipRect.x), (int)(pcmd->ClipRect.w - pcmd->ClipRect.y));
				  glDrawElements(GL_TRIANGLES, (GLsizei)pcmd->ElemCount, sizeof(ImDrawIdx) == 2 ? GL_UNSIGNED_SHORT : GL_UNSIGNED_INT, idx_buffer);
			  }
			  idx_buffer += pcmd->ElemCount;
		  }
	  }
#undef OFFSETOF

	  // Restore modified state
	  glDisableClientState(GL_COLOR_ARRAY);
	  glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	  glDisableClientState(GL_VERTEX_ARRAY);
	  glBindTexture(GL_TEXTURE_2D, last_texture);
	  glMatrixMode(GL_MODELVIEW);
	  glPopMatrix();
	  glMatrixMode(GL_PROJECTION);
	  glPopMatrix();
	  glPopAttrib();
	  glViewport(last_viewport[0], last_viewport[1], (GLsizei)last_viewport[2], (GLsizei)last_viewport[3]);
  }

  bool ImGui_ImplGlfw_CreateDeviceObjects()
  {
	  ImGuiIO& io = ImGui::GetIO();

	  // Build texture atlas
	  unsigned char* pixels;
	  int width, height;
	  io.Fonts->GetTexDataAsAlpha8(&pixels, &width, &height);

	  // Create OpenGL texture
	  GLint last_texture;
	  glGetIntegerv(GL_TEXTURE_BINDING_2D, &last_texture);
	  glGenTextures(1, &g_FontTexture);
	  glBindTexture(GL_TEXTURE_2D, g_FontTexture);
	  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	  glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA, width, height, 0, GL_ALPHA, GL_UNSIGNED_BYTE, pixels);

	  // Store our identifier
	  io.Fonts->TexID = (void *)(intptr_t)g_FontTexture;

	  // Cleanup (don't clear the input data if you want to append new fonts later)
	  io.Fonts->ClearInputData();
	  io.Fonts->ClearTexData();
	  glBindTexture(GL_TEXTURE_2D, last_texture);

	  return true;
  }

  void ImGui_ImplGlfw_NewFrame()
  {
	  if (!g_FontTexture)
		  ImGui_ImplGlfw_CreateDeviceObjects();

	  ImGuiIO& io = ImGui::GetIO();

	  // Setup display size (every frame to accommodate for window resizing)
	  int w, h;
	  int display_w, display_h;

	  w = glutGet(GLUT_WINDOW_WIDTH);
	  h = glutGet(GLUT_WINDOW_HEIGHT);

	  display_w = w;
	  display_h = h;

	  io.DisplaySize = ImVec2((float)w, (float)h);
	  io.DisplayFramebufferScale = ImVec2((float)display_w / w, (float)display_h / h);

	  // Setup time step
	  double current_time = glutGet(GLUT_ELAPSED_TIME); // glfwGetTime();
	  io.DeltaTime = g_Time > 0.0 ? (float)(current_time - g_Time) : (float)(1.0f / 60.0f);
	  g_Time = current_time;

	  // Setup inputs
	  // (we already got mouse wheel, keyboard keys & characters from glfw callbacks polled in glfwPollEvents())
	  if (true) // (glfwGetWindowAttrib(g_Window, GLFW_FOCUSED))
	  {
		  io.MousePos = ImVec2((float)g_cursorX, (float)g_cursorY);   // Mouse position in screen coordinates (set to -1,-1 if no mouse / on another screen, etc.)
	  }
	  else
	  {
		  io.MousePos = ImVec2(-1, -1);
	  }

	  for (int i = 0; i < 3; i++)
	  {
		  io.MouseDown[i] = g_MousePressed[i];    // If a mouse press event came, always pass it as "mouse held this frame", so we don't miss click-release events that are shorter than 1 frame.
		  g_MousePressed[i] = false;
	  }

	  // Start the frame
	  ImGui::NewFrame();
  }


  bool    ImGui_ImplGlfw_Init()
  {
	  ImGuiIO& io = ImGui::GetIO();
	  io.KeyMap[ImGuiKey_Tab] = 9;                     // Keyboard mapping. ImGui will use those indices to peek into the io.KeyDown[] array.
	  io.KeyMap[ImGuiKey_LeftArrow] = GLUT_KEY_LEFT;
	  io.KeyMap[ImGuiKey_RightArrow] = GLUT_KEY_RIGHT;
	  io.KeyMap[ImGuiKey_UpArrow] = GLUT_KEY_UP;
	  io.KeyMap[ImGuiKey_DownArrow] = GLUT_KEY_DOWN;
	  io.KeyMap[ImGuiKey_PageUp] = GLUT_KEY_PAGE_UP;
	  io.KeyMap[ImGuiKey_PageDown] = GLUT_KEY_PAGE_DOWN;
	  io.KeyMap[ImGuiKey_Home] = GLUT_KEY_HOME;
	  io.KeyMap[ImGuiKey_End] = GLUT_KEY_END;
	  io.KeyMap[ImGuiKey_Delete] = 127;
	  io.KeyMap[ImGuiKey_Backspace] = 8;
	  io.KeyMap[ImGuiKey_Enter] = 13;
	  io.KeyMap[ImGuiKey_Escape] = 27;
	  io.KeyMap[ImGuiKey_A] = 65;
	  io.KeyMap[ImGuiKey_C] = 67;
	  io.KeyMap[ImGuiKey_V] = 86;
	  io.KeyMap[ImGuiKey_X] = 88;
	  io.KeyMap[ImGuiKey_Y] = 89;
	  io.KeyMap[ImGuiKey_Z] = 90;

	  io.RenderDrawListsFn = ImGui_ImplGlfw_RenderDrawLists;      // Alternatively you can set this to NULL and call ImGui::GetDrawData() after ImGui::Render() to get the same ImDrawData pointer.
	  io.SetClipboardTextFn = NULL; // ImGui_ImplGlfw_SetClipboardText;
	  io.GetClipboardTextFn = NULL; // ImGui_ImplGlfw_GetClipboardText;
#ifdef _WIN32
									//io.ImeWindowHandle = glfwGetWin32Window(g_Window);
#endif

									//glfwSetKeyCallback(window, ImGui_ImplGlFw_KeyCallback);
									//glfwSetCharCallback(window, ImGui_ImplGlfw_CharCallback);

	  return true;
  }

  void ImGui_ImplGlfw_InvalidateDeviceObjects()
  {
	  if (g_FontTexture)
	  {
		  glDeleteTextures(1, &g_FontTexture);
		  ImGui::GetIO().Fonts->TexID = 0;
		  g_FontTexture = 0;
	  }
  }

  void ImGui_ImplGlfw_Shutdown()
  {
	  ImGui_ImplGlfw_InvalidateDeviceObjects();
	  ImGui::Shutdown();
  }

  // ============================== END IMGUI   ==============================


  /* initialize GLUT */
  void initWindowState(int& argc, char** argv, Renderer* rend, RenderParam* rp, 
	  const std::string name, const size_t width, const size_t height, const bool fullscreen, const bool mouseMode)
  {
    g_width = width;
    g_height = height;
	g_pRenderer = rend; // Local pointer
	g_pRenderParam = rp; // Local pointer
	g_pRenderer->resize(g_width, g_height);

    g_fullscreen = fullscreen;
    flip14 = mouseMode;
    glutInit(&argc, argv);
    glutInitWindowSize((GLsizei)g_width, (GLsizei)g_height);
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
    glutInitWindowPosition(0, 0);
    g_window = glutCreateWindow(name.c_str());

	// Setup ImGui binding
	ImGui_ImplGlfw_Init();

    if (g_fullscreen) glutFullScreen();
    glutDisplayFunc(displayFunc);
    glutIdleFunc(idleFunc);
    glutKeyboardFunc(keyboardFunc);
    glutSpecialFunc(specialFunc);
    glutMouseFunc(clickFunc);
	glutPassiveMotionFunc(cursorPos);
    glutMotionFunc(motionFunc);
    glutReshapeFunc(reshapeFunc);

	ImGui_ImplGlfw_Shutdown();
  }

}

