// ======================================================================== //
// Copyright 2009-2016 Intel Corporation                                    //
//                                                                          //
// Licensed under the Apache License, Version 2.0 (the "License");          //
// you may not use this file except in compliance with the License.         //
// You may obtain a copy of the License at                                  //
//                                                                          //
//     http://www.apache.org/licenses/LICENSE-2.0                           //
//                                                                          //
// Unless required by applicable law or agreed to in writing, software      //
// distributed under the License is distributed on an "AS IS" BASIS,        //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. //
// See the License for the specific language governing permissions and      //
// limitations under the License.                                           //
//																			//
// Modified by Jacob Munkberg												//
// ======================================================================== //

#pragma once

#include "../default.h"
#include "differentialgeometry.h"
#include "shapesampler.h"

namespace embree
{  
	class Light {
	public:
		virtual Vec3fa eval(const Vec3fa& wo) const = 0;
		virtual Vec3fa sample(const DifferentialGeometry& dg, Sample3f& wi, float& tMax, const Vec2f& s) const = 0;
		virtual void transform(const AffineSpace3fa& space) = 0;
		virtual bool isEnvLight() { return false; }
	};

  class AmbientLight : public Light
  {
  public:
    AmbientLight () {}
    
    AmbientLight (const Vec3fa& L) 
    : L(L) {}
   
    inline virtual void transform(const AffineSpace3fa& space) {}
    
	inline virtual Vec3fa eval(const Vec3fa& wo) const { return Vec3fa(L); }

	inline virtual Vec3fa sample(const DifferentialGeometry& dg, Sample3f& wi, float& tMax, const Vec2f& s) const 
	{
		wi = cosineSampleHemisphere(s.x, s.y, dg.Ns);
		tMax = 1e20f;
		return Vec3fa(L);
	}

	virtual bool isEnvLight() { return true; }

  public:
    Vec3fa L;                  //!< radiance of ambient light
  };
  
  struct PointLight : public Light
  {
  public:
    PointLight() {}
    
    PointLight (const Vec3fa& P, const Vec3fa& I) 
    : P(P), I(I) {}
    
    inline virtual void transform(const AffineSpace3fa& space) {
		P = xfmPoint(space,P);
    }

	inline virtual Vec3fa eval(const Vec3fa& wo) const { return Vec3fa(I); }

	inline virtual Vec3fa sample(const DifferentialGeometry& dg,
		Sample3f& wi,
		float& tMax,
		const Vec2f& s) const 
	{
		Vec3fa d = P - dg.P;
		float distance = length(d);
		wi = Sample3f(d*rcp(distance), distance*distance);
		tMax = distance;
		return I;
	}
    
  public:
    Vec3fa P;                  //!< position of point light
    Vec3fa I;                  //!< radiant intensity of point light
  };
  
  struct DirectionalLight : public Light
  {
  public:
    DirectionalLight () {}
    
    DirectionalLight (const Vec3fa& D, const Vec3fa& E) 
    : D(D), E(E) {}
    
    inline virtual void transform(const AffineSpace3fa& space) {
      D = xfmVector(space,D);
    }

	inline virtual Vec3fa eval(const Vec3fa& wo) const { return Vec3fa(E); }
    
	inline virtual Vec3fa sample(const DifferentialGeometry& dg,
		Sample3f& wi,
		float& tMax,
		const Vec2f& s) const
	{
		wi = Sample3f(neg(normalize(D)), 1.0f);
		tMax = inf;
		return E;
	}

  public:
    Vec3fa D;                  //!< Light direction
    Vec3fa E;                  //!< Irradiance (W/m^2)
  };
  
  struct SpotLight : public Light
  {
    SpotLight () {}
    
    SpotLight (const Vec3fa& P, const Vec3fa& D, const Vec3fa& I, const float angleMin, const float angleMax)
    : P(P), D(D), I(I), angleMin(angleMin), angleMax(angleMax) {}
    
	inline virtual Vec3fa eval(const Vec3fa& wo) const { return Vec3fa(I); }

	inline virtual Vec3fa sample(const DifferentialGeometry& dg,
		Sample3f& wi,
		float& tMax,
		const Vec2f& s) const
	{
		assert(0);
		return Vec3fa(I);
	}

	inline virtual void transform(const AffineSpace3fa& space) {
		P = xfmPoint(space, P);
		D = xfmVector(space, D);
    }
    
  public:
    Vec3fa P;                 //!< Position of the spot light
    Vec3fa D;                 //!< Light direction of the spot light
    Vec3fa I;                 //!< Radiant intensity (W/sr)
    float angleMin, angleMax; //!< Linear falloff region
  };
  
  struct DistantLight : public Light
  {
  public:
    DistantLight () {}
    
    DistantLight (const Vec3fa& D, const Vec3fa& L, const float halfAngle) 
    : D(D), L(L), halfAngle(halfAngle), radHalfAngle(deg2rad(halfAngle)), cosHalfAngle(cos(deg2rad(halfAngle))) {}
    
	inline virtual void transform(const AffineSpace3fa& space) { D = xfmVector(space, D); }

	inline Vec3fa eval(const Vec3fa& wo) const { return (-dot(wo, D) >= cosHalfAngle) ? L : Vec3fa(0.0f); }

	inline Vec3fa sample(const DifferentialGeometry& dg,
		Sample3f& wi,
		float& tMax,
		const Vec2f& s) const
	{
		wi = UniformSampleCone(s.x, s.y, radHalfAngle, neg(D));
		tMax = 1e20f;
		return L;
	}
    
  public:
    Vec3fa D;            //!< Light direction
    Vec3fa L;            //!< Radiance (W/(m^2*sr))
    float halfAngle;     //!< Half illumination angle
    float radHalfAngle;  //!< Half illumination angle in radians
    float cosHalfAngle;  //!< Cosine of half illumination angle
  };
  
  struct TriangleLight : public Light
  {
    TriangleLight() {}
    
    TriangleLight (const Vec3fa& v0, const Vec3fa& v1, const Vec3fa& v2, const Vec3fa& L) 
    : v0(v0), v1(v1), v2(v2), L(L) {}
    
	inline virtual void transform(const AffineSpace3fa& space) {
		v0 = xfmPoint(space, v0);
		v1 = xfmPoint(space, v1);
		v2 = xfmPoint(space, v2);
    }

	inline Vec3fa eval(const Vec3fa& wo) const { 
		assert(0);
		return Vec3fa(0.0f); 
	}

	inline Vec3fa sample(const DifferentialGeometry& dg,
		Sample3f& wi,
		float& tMax,
		const Vec2f& s) const
	{
		assert(0);
		return L;
	}

  public:
    Vec3fa v0;
    Vec3fa v1;
    Vec3fa v2;
    Vec3fa L;
  };
  
  struct QuadLight : public Light
  {
    QuadLight() {}
    
    QuadLight (const Vec3fa& v0, const Vec3fa& v1, const Vec3fa& v2, const Vec3fa& v3, const Vec3fa& L) 
    : v0(v0), v1(v1), v2(v2), v3(v3), L(L) {}
    
	inline virtual void transform(const AffineSpace3fa& space) {
		v0 = xfmPoint(space, v0);
		v1 = xfmPoint(space, v1);
		v2 = xfmPoint(space, v2);
		v3 = xfmPoint(space, v3);
    }

	inline Vec3fa eval(const Vec3fa& wo) const {
		assert(0);
		return Vec3fa(0.0f);
	}

	inline Vec3fa sample(const DifferentialGeometry& dg,
		Sample3f& wi,
		float& tMax,
		const Vec2f& s) const
	{
		assert(0);
		return L;
	}

    
  public:
    Vec3fa v0;
    Vec3fa v1;
    Vec3fa v2;
    Vec3fa v3;
    Vec3fa L;
  };
}
