// ======================================================================== //
// Copyright 2009-2016 Intel Corporation                                    //
//                                                                          //
// Licensed under the Apache License, Version 2.0 (the "License");          //
// you may not use this file except in compliance with the License.         //
// You may obtain a copy of the License at                                  //
//                                                                          //
//     http://www.apache.org/licenses/LICENSE-2.0                           //
//                                                                          //
// Unless required by applicable law or agreed to in writing, software      //
// distributed under the License is distributed on an "AS IS" BASIS,        //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. //
// See the License for the specific language governing permissions and      //
// limitations under the License.                                           //
//																			//
// Modified by Jacob Munkberg												//
// ======================================================================== //

#pragma once

#include "../default.h"
#include "brdf.h"
#include "optics.h"
#include "texture.h"

namespace embree
{
  enum MaterialTy 
  {
    MATERIAL_MATTE, 
	MATERIAL_TEXTURED_MATTE,
  };

  /* Thread-specific scratch pad for texture lookups at a certain hit */
  struct ShaderContext {
	  Vec3fa tex[1]; // TODO: Make this larger when needed
  };

  // Base material class
  class Material {
  public:
	  Material() {}

	  Material(MaterialTy type) : ty(type) {}

	  virtual void preprocess(const Vec3fa& wo, const DifferentialGeometry& dg, ShaderContext& context) {};

	  virtual Vec3fa eval(const ShaderContext& context, const Vec3fa& wo, const DifferentialGeometry& dg, const Vec3fa& wi, float& pdf) = 0;

	  virtual Vec3fa albedo(const ShaderContext& context) = 0;

	  virtual Vec3fa sample(const ShaderContext& context, const Vec3fa& Lw, const Vec3fa& wo, const DifferentialGeometry& dg, 
							Sample3f& wi_o, const Vec2f& s) = 0;
  public:
	  int ty;
  };

  class MatteMaterial : public Material
  {
  public:
    MatteMaterial (const Vec3fa& reflectance)
    : Material(MATERIAL_MATTE), lambert(reflectance) {}

	virtual Vec3fa eval(const ShaderContext& context, const Vec3fa& wo, const DifferentialGeometry& dg, const Vec3fa& wi, float& pdf) override
	{
		pdf = max(0.0f, dot(dg.Ns, wi))*float(one_over_pi);
		return lambert.eval(wo, dg, wi);
	}

	virtual Vec3fa sample(const ShaderContext& context, const Vec3fa& Lw, const Vec3fa& wo, const DifferentialGeometry& dg, Sample3f& wi_o, const Vec2f& s) override
	{
		return lambert.sample(wo, dg, wi_o, s);
	}

	virtual Vec3fa albedo(const ShaderContext& context) override { return lambert.R; }

  public:
	Lambertian lambert;
  };
 
  class TexturedMatteMaterial : public Material
  {
  public:
	  TexturedMatteMaterial(const Texture* _map_Kd)
		  : Material(MATERIAL_TEXTURED_MATTE), map_Kd(_map_Kd), lambert(Vec3fa(1,1,1)), Kd(Vec3fa(1,1,1)) {}

	  void preprocess(const Vec3fa& wo, const DifferentialGeometry& dg, ShaderContext& context) override
	  {
		  context.tex[0] = map_Kd ? map_Kd->getTexel3f_bilinear(dg.u, dg.v) : Vec3fa(1.0f);
	  }

	  virtual Vec3fa eval(const ShaderContext& context, const Vec3fa& wo, const DifferentialGeometry& dg, const Vec3fa& wi, float& pdf) override
	  {
		  pdf = max(0.0f, dot(dg.Ns, wi))*float(one_over_pi);
		  return context.tex[0] * lambert.eval(wo, dg, wi);
	  }

	  virtual Vec3fa sample(const ShaderContext& context, const Vec3fa& Lw, const Vec3fa& wo, const DifferentialGeometry& dg, Sample3f& wi_o, const Vec2f& s) override
	  {
		  return context.tex[0] *lambert.sample(wo, dg, wi_o, s);
	  }

	  virtual Vec3fa albedo(const ShaderContext& context) override { return context.tex[0]; }

  public:
	  Lambertian lambert;
	  const Texture* map_Kd;	/*< Kd texture */
	  Vec3fa Kd;
  };
}
