// ======================================================================== //
// Copyright 2009-2016 Intel Corporation                                    //
//                                                                          //
// Licensed under the Apache License, Version 2.0 (the "License");          //
// you may not use this file except in compliance with the License.         //
// You may obtain a copy of the License at                                  //
//                                                                          //
//     http://www.apache.org/licenses/LICENSE-2.0                           //
//                                                                          //
// Unless required by applicable law or agreed to in writing, software      //
// distributed under the License is distributed on an "AS IS" BASIS,        //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. //
// See the License for the specific language governing permissions and      //
// limitations under the License.                                           //
//																			//
// Modified by Jacob Munkberg												//
// ======================================================================== //

#include "scene.h"

namespace embree
{
  struct SceneGraphConverter
  {
    Scene* scene;
    std::map<Ref<SceneGraph::MaterialNode>, size_t> material2id;
    std::map<Ref<SceneGraph::Node>, size_t> geometry2id;
    
    SceneGraphConverter (Ref<SceneGraph::Node> in, Scene* scene, Scene::InstancingMode instancing)
      : scene(scene)
    { 
      convertLights(in,one,one);

      if (instancing != Scene::INSTANCING_NONE) 
      {
        if (instancing == Scene::INSTANCING_SCENE_GROUP) 
        {
          in->reset();
          in->calculateInDegree();
          in->calculateClosed();
        }
        convertInstances(in,one,one);
      }
      else
        convertGeometries(scene->geometries,in,one,one);
    }

    int convert(Ref<SceneGraph::MaterialNode> node) 
    {
      if (material2id.find(node) == material2id.end()) {
        scene->materials.push_back(node->material);
        material2id[node] = scene->materials.size()-1;
      }
      return material2id[node];
    }

    Ref<Scene::Geometry> convertTriangleMesh(Ref<SceneGraph::TriangleMeshNode> mesh, const AffineSpace3fa& space0, const AffineSpace3fa& space1)
    {
      int materialID = convert(mesh->material);
      
      Scene::TriangleMesh* objmesh = new Scene::TriangleMesh();
      const LinearSpace3fa nspace0 = rcp(space0.l).transposed();
      objmesh->v. resize(mesh->v. size()); for (size_t i=0; i<mesh->v. size(); i++) objmesh->v [i] = xfmPoint ( space0,mesh->v [i]);
      objmesh->v2.resize(mesh->v2.size()); for (size_t i=0; i<mesh->v2.size(); i++) objmesh->v2[i] = xfmPoint ( space1,mesh->v2[i]);
      objmesh->vn.resize(mesh->vn.size()); for (size_t i=0; i<mesh->vn.size(); i++) objmesh->vn[i] = xfmVector(nspace0,mesh->vn[i]);
      objmesh->vt = mesh->vt;
      
      objmesh->triangles.resize(mesh->triangles.size());
      for (size_t i=0; i<mesh->triangles.size(); i++) {
        SceneGraph::TriangleMeshNode::Triangle& tri = mesh->triangles[i];
        objmesh->triangles[i] = Scene::Triangle(tri.v0,tri.v1,tri.v2,materialID);
      }
      objmesh->meshMaterialID = materialID;
      return objmesh;
    }

    Ref<Scene::Geometry> convertQuadMesh(Ref<SceneGraph::QuadMeshNode> mesh, const AffineSpace3fa& space0, const AffineSpace3fa& space1)
    {
      int materialID = convert(mesh->material);
      
      Scene::QuadMesh* objmesh = new Scene::QuadMesh();
      const LinearSpace3fa nspace0 = rcp(space0.l).transposed();
      objmesh->v. resize(mesh->v. size()); for (size_t i=0; i<mesh->v. size(); i++) objmesh->v [i] = xfmPoint ( space0,mesh->v [i]);
      objmesh->v2.resize(mesh->v2.size()); for (size_t i=0; i<mesh->v2.size(); i++) objmesh->v2[i] = xfmPoint ( space1,mesh->v2[i]);
      objmesh->vn.resize(mesh->vn.size()); for (size_t i=0; i<mesh->vn.size(); i++) objmesh->vn[i] = xfmVector(nspace0,mesh->vn[i]);
      objmesh->vt = mesh->vt;
      
      objmesh->quads.resize(mesh->quads.size());
      for (size_t i=0; i<mesh->quads.size(); i++) {
        SceneGraph::QuadMeshNode::Quad& quad = mesh->quads[i];
        objmesh->quads[i] = Scene::Quad(quad.v0,quad.v1,quad.v2,quad.v3);
      }
      objmesh->meshMaterialID = materialID;
      return objmesh;
    }

    Ref<Scene::Geometry> convertSubdivMesh(Ref<SceneGraph::SubdivMeshNode> mesh, const AffineSpace3fa& space0, const AffineSpace3fa& space1)
    {
      int materialID = convert(mesh->material);
      
      Scene::SubdivMesh* subdivmesh = new Scene::SubdivMesh();
      const LinearSpace3fa nspace0 = rcp(space0.l).transposed();
      
      subdivmesh->positions.resize(mesh->positions.size()); 
      for (size_t i=0; i<mesh->positions.size(); i++) 
        subdivmesh->positions[i] = xfmPoint(space0,mesh->positions[i]);
      
      subdivmesh->normals.resize(mesh->normals.size()); 
      for (size_t i=0; i<mesh->normals.size(); i++) 
        subdivmesh->normals[i] = xfmVector(nspace0,mesh->normals[i]);

      subdivmesh->texcoords = mesh->texcoords;
      subdivmesh->position_indices = mesh->position_indices;
      subdivmesh->normal_indices = mesh->normal_indices;
      subdivmesh->texcoord_indices = mesh->texcoord_indices;
      subdivmesh->verticesPerFace = mesh->verticesPerFace;
      subdivmesh->holes = mesh->holes;
      subdivmesh->edge_creases = mesh->edge_creases;
      subdivmesh->edge_crease_weights = mesh->edge_crease_weights;
      subdivmesh->vertex_creases = mesh->vertex_creases;
      subdivmesh->vertex_crease_weights = mesh->vertex_crease_weights;
      subdivmesh->materialID = materialID;

	  subdivmesh->numVertices = mesh->positions.size();
	  subdivmesh->numFaces = mesh->verticesPerFace.size();
	  subdivmesh->numEdges = mesh->position_indices.size();
	  subdivmesh->numEdgeCreases = mesh->edge_creases.size();
	  subdivmesh->numVertexCreases = mesh->vertex_creases.size();
	  subdivmesh->numHoles = mesh->holes.size();
	  subdivmesh->geomID = -1;
	  subdivmesh->subdivlevel  = new float[subdivmesh->numEdges];
	  subdivmesh->face_offsets = new int[subdivmesh->numFaces];

	  for (size_t i = 0; i<subdivmesh->numEdges; i++) subdivmesh->subdivlevel[i] = 1.0f;
	  int offset = 0;
	  for (size_t i = 0; i<subdivmesh->numFaces; i++)
	  {
		  subdivmesh->face_offsets[i] = offset;
		  offset += mesh->verticesPerFace[i];
	  }
	  subdivmesh->face_neighbors = mesh->face_neighbors;
	  subdivmesh->edge_neighbors = mesh->edge_neighbors;

      return subdivmesh;
    }

     void convertLights(Ref<SceneGraph::Node> node, const AffineSpace3fa& space0, const AffineSpace3fa& space1)
    {
      if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
        convertLights(xfmNode->child, space0*xfmNode->xfm0, space1*xfmNode->xfm1);
      } 
      else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>()) {
        for (auto child : groupNode->children) convertLights(child,space0,space1);
      }
	  else if ( Ref<SceneGraph::LightNode> light = node.dynamicCast<SceneGraph::LightNode>() ){
		  light->light->transform(space0);
		  scene->lights.push_back(light->light);
	  }
    }

    void convertGeometries(std::vector<Ref<Scene::Geometry>>& group, Ref<SceneGraph::Node> node, const AffineSpace3fa& space0, const AffineSpace3fa& space1)
    {
      if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
        convertGeometries(group,xfmNode->child, space0*xfmNode->xfm0, space1*xfmNode->xfm1);
      } 
      else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>()) {
        for (auto child : groupNode->children) convertGeometries(group,child,space0,space1);
      }
      else if (Ref<SceneGraph::TriangleMeshNode> mesh = node.dynamicCast<SceneGraph::TriangleMeshNode>()) {
        group.push_back(convertTriangleMesh(mesh,space0,space1));
      }
      else if (Ref<SceneGraph::QuadMeshNode> mesh = node.dynamicCast<SceneGraph::QuadMeshNode>()) {
        group.push_back(convertQuadMesh(mesh,space0,space1));
      }
      else if (Ref<SceneGraph::SubdivMeshNode> mesh = node.dynamicCast<SceneGraph::SubdivMeshNode>()) {
        group.push_back(convertSubdivMesh(mesh,space0,space1));
      }
    }

    int lookupGeometries(Ref<SceneGraph::Node> node)
    {
      if (geometry2id.find(node) == geometry2id.end())
      {
        std::vector<Ref<Scene::Geometry>> geometries;
        convertGeometries(geometries,node,one,one);
        
        if (geometries.size() == 1)
          scene->geometries.push_back(geometries[0]);
        else 
          scene->geometries.push_back(new Scene::Group(geometries));
        
        geometry2id[node] = scene->geometries.size()-1;
      }
      return geometry2id[node];
    }

    void convertInstances(Ref<SceneGraph::Node> node, const AffineSpace3fa& space0, const AffineSpace3fa& space1)
    {
      if (node->isClosed()) {
        scene->geometries.push_back(new Scene::Instance(space0,lookupGeometries(node)));
      }
      else if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
        convertInstances(xfmNode->child, space0*xfmNode->xfm0, space1*xfmNode->xfm1);
      } 
      else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>()) {
        for (auto child : groupNode->children) convertInstances(child,space0,space1);
      }
    }
  };

  void Scene::add(Ref<SceneGraph::Node> node, Scene::InstancingMode instancing) {
    SceneGraphConverter(node,this,instancing);
  }
};
