// ======================================================================== //
// Copyright 2009-2016 Intel Corporation                                    //
//                                                                          //
// Licensed under the Apache License, Version 2.0 (the "License");          //
// you may not use this file except in compliance with the License.         //
// You may obtain a copy of the License at                                  //
//                                                                          //
//     http://www.apache.org/licenses/LICENSE-2.0                           //
//                                                                          //
// Unless required by applicable law or agreed to in writing, software      //
// distributed under the License is distributed on an "AS IS" BASIS,        //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. //
// See the License for the specific language governing permissions and      //
// limitations under the License.                                           //
//																			//
// Modified by Jacob Munkberg												//
// ======================================================================== //

#pragma once

#include "../default.h"
#include "../scenegraph/scenegraph.h"

namespace embree
{
  /*! Scene representing the OBJ file */
  struct Scene
  {
    Scene () {}

    enum InstancingMode { INSTANCING_NONE, INSTANCING_GEOMETRY, INSTANCING_SCENE_GEOMETRY, INSTANCING_SCENE_GROUP };
    void add (Ref<SceneGraph::Node> node, InstancingMode instancing);

    /*! OBJ Triangle */
    struct Triangle 
    {
    public:
      Triangle () {}
      
      Triangle (const Triangle& other) 
      : v0(other.v0), v1(other.v1), v2(other.v2), materialID(other.materialID) {}

      Triangle (int v0, int v1, int v2, int materialID) 
      : v0(v0), v1(v1), v2(v2), materialID(materialID) {}

    public:
      int v0, v1, v2, materialID;
    };

    /*! OBJ Quad */
    struct Quad 
    {
    public:
      Quad () {}

      Quad (int v0, int v1, int v2, int v3) 
      : v0(v0), v1(v1), v2(v2), v3(v3) {}

    public:
      int v0, v1, v2, v3;
    };

    struct Geometry : public RefCount
    {
      enum Type { TRIANGLE_MESH, SUBDIV_MESH, HAIR_SET, INSTANCE, GROUP, QUAD_MESH, LINE_SEGMENTS };
      Type type;

      Geometry (Type type) : type(type) {}
    };

    /*! Triangle Mesh. */
    struct TriangleMesh : public Geometry
    {
      TriangleMesh () : Geometry(TRIANGLE_MESH) {}
      avector<Vec3fa> v;
      avector<Vec3fa> v2;
      avector<Vec3fa> vn;
      std::vector<Vec2f> vt;
      std::vector<Triangle> triangles;
      std::vector<Quad> quads; // FIXME: remove
      int meshMaterialID;
	  int geomID;
    };

    /*! Quad Mesh. */
    struct QuadMesh : public Geometry
    {
      QuadMesh () : Geometry(QUAD_MESH) {}
      avector<Vec3fa> v;
      avector<Vec3fa> v2;
      avector<Vec3fa> vn;
      std::vector<Vec2f> vt;
      std::vector<Quad> quads;
      int meshMaterialID;
	  int geomID;
    };

    /*! Subdivision Mesh. */
    struct SubdivMesh : public Geometry
    {
      SubdivMesh () : Geometry(SUBDIV_MESH) {}

	  SubdivMesh(Ref<Scene::SubdivMesh> in) : Geometry(SUBDIV_MESH)
	  {
		  // Copy over data ...
		  positions = in->positions;
		  normals = in->normals;
		  texcoords = in->texcoords;
		  position_indices = in->position_indices;
		  normal_indices = in->normal_indices;
		  texcoord_indices = in->texcoord_indices;
		  verticesPerFace = in->verticesPerFace;
		  holes = in->holes;
		  edge_creases = in->edge_creases;
		  edge_crease_weights = in->edge_crease_weights;
		  vertex_creases = in->vertex_creases;
		  vertex_crease_weights = in->vertex_crease_weights;
		  materialID = in->materialID;

		  numVertices = in->positions.size();
		  numFaces = in->verticesPerFace.size();
		  numEdges = in->position_indices.size();
		  numEdgeCreases = in->edge_creases.size();
		  numVertexCreases = in->vertex_creases.size();
		  numHoles = in->holes.size();
		  geomID = in->geomID;
		  subdivlevel = new float[numEdges];
		  face_offsets = new int[numFaces];
		  face_neighbors = in->face_neighbors;
		  edge_neighbors = in->edge_neighbors;

		  for (size_t i = 0; i<numEdges; i++) subdivlevel[i] = in->subdivlevel[i];
		  for (size_t i = 0; i<numFaces; i++) face_offsets[i] = in->face_offsets[i];
	  }

      avector<Vec3fa> positions;                //!< vertex positions
      avector<Vec3fa> normals;                  //!< face vertex normals
      std::vector<Vec2f> texcoords;             //!< face texture coordinates
      std::vector<int> position_indices;        //!< position indices for all faces
      std::vector<int> normal_indices;          //!< normal indices for all faces
      std::vector<int> texcoord_indices;        //!< texcoord indices for all faces
      std::vector<int> verticesPerFace;         //!< number of indices of each face
      std::vector<int> holes;                   //!< face ID of holes
	  float* subdivlevel;						//!< subdivision level
      std::vector<Vec2i> edge_creases;          //!< index pairs for edge crease 
      std::vector<float> edge_crease_weights;   //!< weight for each edge crease
      std::vector<int> vertex_creases;          //!< indices of vertex creases
      std::vector<float> vertex_crease_weights; //!< weight for each vertex crease
	  int *face_offsets;
	  int numVertices;
	  int numFaces;
	  int numEdges;
	  int numEdgeCreases;
	  int numVertexCreases;
	  int numHoles;
	  int materialID;
	  int geomID;

	  std::vector<int> face_neighbors;		//!< Face neighbors 
	  std::vector<int> edge_neighbors;		//!< Edge neighbors
    };

	struct SubdivMeshKeyFrame
	{
		SubdivMesh** subdiv;       //!< list of subdiv meshes
		int numSubdivMeshes;       //!< number of subdiv meshes
	};

    struct Instance : public Geometry
    {
      ALIGNED_STRUCT;

      Instance(const AffineSpace3fa& space, int geomID)
        : Geometry(INSTANCE), space(space), geomID(geomID) {}

    public:
      AffineSpace3fa space;
      int geomID;
    };

    struct Group : public Geometry
    {
      Group (std::vector<Ref<Geometry>> children)
        : Geometry(GROUP), children(children) {}

      size_t size() const { return children.size(); }
      Ref<Geometry> at(size_t i) { return children[i]; }

    public:
      std::vector<Ref<Geometry>> children;
    };

    bool empty() const {
      return geometries.size() == 0;
    }

  public:
    avector<Material*> materials;                  //!< material list
    std::vector<Ref<Geometry> > geometries;        //!< list of geometries
	std::vector<Light*> lights;					   //!< list of lights: Todo - vector or avector?		

	SubdivMeshKeyFrame** subdivMeshKeyFrames;
	int numSubdivMeshKeyFrames;
  };

  typedef Scene::Geometry Geometry;
  typedef Scene::TriangleMesh TriangleMesh;
  typedef Scene::QuadMesh QuadMesh;
  typedef Scene::SubdivMesh SubdivMesh;
  typedef Scene::SubdivMeshKeyFrame SubdivMeshKeyFrame;
  typedef Scene::Instance Instance;
  typedef Scene::Group Group;
  typedef Scene::Triangle Triangle;
  typedef Scene::Quad Quad;

}
