// ======================================================================== //
// Copyright 2009-2016 Intel Corporation                                    //
//                                                                          //
// Licensed under the Apache License, Version 2.0 (the "License");          //
// you may not use this file except in compliance with the License.         //
// You may obtain a copy of the License at                                  //
//                                                                          //
//     http://www.apache.org/licenses/LICENSE-2.0                           //
//                                                                          //
// Unless required by applicable law or agreed to in writing, software      //
// distributed under the License is distributed on an "AS IS" BASIS,        //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. //
// See the License for the specific language governing permissions and      //
// limitations under the License.                                           //
//																			//
// Modified by Jacob Munkberg												//
// ======================================================================== //

#include "texture.h"
//#include "..\..\..\common\sys\platform.h"

namespace embree
{
  bool isPowerOf2 (unsigned int x)
  {
    while (((x % 2) == 0) && x > 1)
      x /= 2;
    return (x == 1);
  }

  std::map<std::string,Texture*> texture_cache;

  Texture::Texture () 
    : width(-1), height(-1), format(INVALID), bytesPerTexel(0), data(nullptr), width_mask(0), height_mask(0) {}
  
  Texture::Texture(Ref<Image> img, const std::string fileName)
    : width(img->width), height(img->height), format(RGBA8), bytesPerTexel(4), data(nullptr), width_mask(0), height_mask(0), fileName(fileName)
  {
    width_mask  = isPowerOf2(width) ? width-1 : 0;
    height_mask = isPowerOf2(height) ? height-1 : 0;

    data = _mm_malloc(4*width*height,64);
    img->convertToRGBA8((unsigned char*)data);
  }

  Texture::Texture (size_t width, size_t height, const Format format, const char* in)
    : width(width), height(height), format(format), bytesPerTexel(getFormatBytesPerTexel(format)), data(nullptr), width_mask(0), height_mask(0)
  {
    width_mask  = isPowerOf2(width) ? width-1 : 0;
    height_mask = isPowerOf2(height) ? height-1 : 0;

    data = _mm_malloc(bytesPerTexel*width*height,64);
    if (in) memcpy(data,in,bytesPerTexel*width*height);
    else    memset(data,0 ,bytesPerTexel*width*height);
  }

  Texture::~Texture () {
    _mm_free(data);
  }

  const char* Texture::format_to_string(const Format format)
  {
    switch (format) {
    case RGBA8  : return "RGBA8";
    case RGB8   : return "RGB8";
    case FLOAT32: return "FLOAT32";
    default     : THROW_RUNTIME_ERROR("invalid texture format");
    }
  }

  Texture::Format Texture::string_to_format(const std::string& str)
  {
    if      (str == "RGBA8") return RGBA8;
    else if (str == "RGB8")  return RGB8;
    else if (str == "FLOAT32") return FLOAT32;
    else THROW_RUNTIME_ERROR("invalid texture format string");
  }

  int Texture::getFormatBytesPerTexel(const Format format)
  {
    switch (format) {
    case RGBA8  : return 4;
    case RGB8   : return 3;
    case FLOAT32: return 4;
    default     : THROW_RUNTIME_ERROR("invalid texture format");
    }
  }

  /*! read png texture from disk */
  Texture* Texture::load(const FileName& fileName)
  {
    if (texture_cache.find(fileName.str()) != texture_cache.end())
      return texture_cache[fileName.str()];

    return texture_cache[fileName.str()] = new Texture(loadImage(fileName),fileName);
  }

  float Texture::getTexel1f(float s, float t) const
  {
	  s = max(s, 0.0f);
	  t = max(t, 0.0f);

	  if (likely(format == Texture::FLOAT32))
	  {
		  const float u = min(max(s, 0.0f), 1.0f);
		  const float v = min(max(t, 0.0f), 1.0f);
		  const int ui = min((int)floorf((width - 1)*u),  width - 1);
		  const int vi = min((int)floorf((height - 1)*v), height - 1);
		  float *p = (float *)data;
		  return p[vi*width + ui];
	  }
	  else if (likely(format == Texture::RGBA8))
	  {
		  int iu = (int)(s * (float)(width));
		  if (width_mask) iu &= width_mask;
		  else                     iu = min(iu, width - 1);
		  int iv = (int)(t * (float)(height));
		  if (height_mask)  iv &= height_mask;
		  else                       iv = min(iv, height - 1);
		  unsigned char* t = (unsigned char*)data + (iv * width + iu) * 4;
		  return (float)t[0] * (1.0f / 255.0f);
	  }
	  return 0.0f;
  }

  Vec3f Texture::getTexel3f(float s, float t) const
  {
	  s = max(s, 0.0f);
	  t = max(t, 0.0f);

	  if (likely(format == Texture::RGBA8))
	  {
		  int iu = (int)(s * (float)(width));
		  if (width_mask)
			  iu &= width_mask;
		  else
			  iu = min(iu, width - 1);
		  int iv = (int)(t * (float)(height));
		  if (height_mask)
			  iv &= height_mask;
		  else
			  iv = min(iv, height - 1);
		  unsigned char *t = (unsigned char*)data + (iv * width + iu) * 4; //texture->bytesPerTexel;
		  return Vec3f((float)t[0] * 1.0f / 255.0f, (float)t[1] * 1.0f / 255.0f, (float)t[2] * 1.0f / 255.0f);
	  }
	  return Vec3f(0.0f);

  }

  Vec3fa Texture::getTexel3f_bilinear(float s, float t) const
  {
	  if (likely(format == Texture::RGBA8))
	  {
		  // SSE-optimized bilinear filtering - corner texel mode
		  __m128 xf = _mm_set1_ps(s*(width - 1));
		  __m128 yf = _mm_set1_ps(t*(height - 1));

		  __m128 xt = _mm_floor_ps(xf);
		  __m128 yt = _mm_floor_ps(yf);

		  __m128 wx = _mm_sub_ps(xf, xt);
		  __m128 wy = _mm_sub_ps(yf, yt);

		  wx = _mm_xor_ps(wx, _mm_set_ps(0.f, -0.f, 0.f, -0.f));
		  wx = _mm_add_ps(wx, _mm_set_ps(0.f, 1.f, 0.f, 1.f));        // (wx, 1-wx, wx, 1-wx)

		  wy = _mm_xor_ps(wy, _mm_set_ps(0.f, 0.f, -0.f, -0.f));
		  wy = _mm_add_ps(wy, _mm_set_ps(0.f, 0.f, 1.f, 1.f));        // (wy, wy, 1-wy, 1-wy)

		  __m128 w = _mm_mul_ps(wx, wy);              // w = bilinear weights

		  __m128i xi = _mm_cvtps_epi32(xt);
		  __m128i yi = _mm_cvtps_epi32(yt);

		  xi = _mm_add_epi32(xi, _mm_set_epi32(1, 0, 1, 0));
		  yi = _mm_add_epi32(yi, _mm_set_epi32(1, 1, 0, 0));

		  xi = _mm_max_epi32(xi, _mm_setzero_si128());
		  yi = _mm_max_epi32(yi, _mm_setzero_si128());

		  xi = _mm_min_epi32(xi, _mm_set1_epi32(width - 1));
		  yi = _mm_min_epi32(yi, _mm_set1_epi32(height - 1));

		  __m128i off = _mm_mullo_epi32(yi, _mm_set1_epi32(width));
		  off = _mm_add_epi32(off, xi);       // off = (y * width + x)
		  off = _mm_mullo_epi32(off, _mm_set1_epi32(bytesPerTexel));     // off = byte offset

		  __m128 c = _mm_setzero_ps();
		  for (int i = 0; i < 4; i++)
		  {
			  const float f = 1.0f / 255.0f;
			  unsigned char *t = (unsigned char*)data + off.m128i_i32[i];
			  Color4 tap = Color4((float)t[0] * f, (float)t[1] * f, (float)t[2] * f, 1.0f);
			  c = _mm_add_ps(c, _mm_mul_ps(tap.m128, _mm_set1_ps(w.m128_f32[i])));
		  }
		  return Vec3fa(c);
	  }
	  return Vec3fa(0.0f);
  }


}
